#ifndef USEVOTE_H
#define USEVOTE_H
/*---------------------------------------------------------------
/* UseVote  (c)1993,94 Ron Dippold All Rights Reserved
/* Version 3.0  - usevote.h
/*
/* This is the general configuration file for all the UseVote
/* modules.  You should be able to use this file for any vote
/* you run once you set it up.
/*
/* Make sure to check every item lest ye be bitten.
/* Stuff for any specific vote goes into usevote.cfg
/*-------------------------------------------------------------*/

/* How many items can we have on a ballot?  20 is insanely high */
/* This saves us from pointers to pointers to pointers */
#define MAXGROUPS 20


/* Name of configuration file for individual votes.  Should be in the */
/* same directory as the rest of the info for the vote */
#define CFGFILE "usevote.cfg"


/* Define here the text that marks the start and end of a message in your
/* mail file or copied mail file.  This will be match only if it is at the
/* start of a line.  For most implementations, it is "From " for the start
/* of a message and nothing for the end of a message (the next "From ")
/* both ends and starts it.   MMDF format uses four ctrl-a's to start
/* _and_ a separate four ctrl-a's to end.  To specify control characters,
/* use the octal (base 8) method, so four ctrl-a's is "\001\001\001\001".
/* If your format doesn't need the end of message, make it all blanks ( "" ),
/* don't comment it out.  The MAILSTART is required.
*/
#define MAILSTART "From "
#define MAILEND   ""


/* ( Only used if you define VOTEACK in usevote.cfg )
/* Place here any commands which you want placed at the start of the
/* "domail" file.  Usually this will be something like "#!/bin/sh\n"
/* to make sure it is run by the "sh" shell.  End every line with a "\n"!
*/
#define VOTEACKPREFIX ""


/* ( Only used if you define VOTEACK in usevote.cfg )
/* Place here any commands which you want placed at the end of the "domail"
/* file.  End every line with a "\n"!
*/
#define VOTEACKSUFFIX ""


/* ( Only used if you define VOTEACK in usevote.cfg )
/* This is the command that will be used to mail each vote ack.  This should
/* work for most Unix systems, but who knows what wierd thing you use?
/*   #s is where the "Subject:" the reader sees goes
/*   #S is #s, but quotes are used around the subject.  PREFERRED.
/*   #v will be replaced with the voter address.        PREFERRED.
/*   #V is #v, but quotes are used around the address.
/*   #f is the name of the file containing the vote ack
/* You can have multiple commands per mailing by separating each command
/* with a "\n".
/*   The "\n sleep 20" has been added because some systems can't handle the
/* rapid pounding of the mail system by UseVote sending out mail.  This makes
/* UseVote pause 20 seconds after sending an ack.  You may increase or decrease this
/* value at will.
/*   The "\rm #f" removes the ack text file after the mail is sent.  "rm"
/* is the Unix shell command for delete, you may have to change it to "erase"
/* or "del" or whatever your system wants.
*/
#define MAILCMD "mail -s #S #v < #f \n sleep 20\n rm #f"


/* ( Only used if you define VOTEACK in usevote.cfg )
/* This #define indicates which characters need to be literal
/* quoted with a backslash (\) for in a person's address for the sake of
/* some systems.  If you don't want any backslashing used, delete everything
/* between the quotes.
*/
#define ADDBADCHAR "!#$%^&*()+{}|'\";:/?[]~`"



/* Define the right margin for vote acknowledgement lists
*/
#define RIGHT 78

/* Page length for interactive mode - 24 is good for most terminals */
#define PAGELEN 24

/* Page width for interactive terminals - 80 is good for most */
#define PAGEWIDE 80

/* Screen clearing code for interactive vote - if you don't know it,
/* "\n\n\n" will print 3 blank lines.  "\f" or "\014" often works.
/* "\033[2J" is the VT100 code.
*/
#define CLRSCRN "\033[2J"


/* Here's where you specify the lines in the header that you think will
/* contain date/time info so uvvote can snag it.  This is prioritized, so
/* those given first will be used first.  You can look at the headers in mail
/* you receive, but the defaults should work for most unix mail implementations.
/* The "Received: from " is the one you would customize for your gateway if
/* you cared to. For me they look like "Received: from qualcomm.com by happy"
*/
#define DTXTS 3                   /* number of entries in datestxt */
#ifdef UVCONFIG
char datestxt[ DTXTS ][ 80 ]= {   /* need at least one entry! */
  "From ",                        /* first line in message, usually */
  "Received: from ",              /* as each gateway gets it... */
  "Date: "                        /* use posting date as last resort */
};
#else
extern char datestxt[ DTXTS ][ 80 ];
#endif


/* Specify accounts/sites which are invalid.  Votes from these will be
/* rejected with an explanitory message.  This can get quite long...
/* I would add your personal address and voting address at the top, in
/* case some idiot uses that as a "Reply-To:"
/* "*" works as a wildcard, "?" works as a character wildcard
*/
#define ATXTS 32                  /* number of entries in addrstxt */
#ifdef UVCONFIG
char addrstxt[ ATXTS ][ 40 ]= {
  "myaccount@*bar.com",           /* your email address */
  "voting@*bar.com",              /* your voting address */
  "*@anon.penet.fi",  "*@anon.sub.net",
  "*admin@*",         "*daemon*@*",
  "anon*@*",          "audit@*",
  "bin@*",            "decnet@*",
  "default@*",        "field@*",
  "guest@*",          "hostmaster@*",
  "install@*",        "mailer*@*",
  "maint@*",          "news@*",
  "newsmaster@*",     "nobody@*",
  "operator@*",       "postmast*@*",
  "root@*",           "shutdown@*",
  "sync@*",           "sys@*",
  "sysop@*",          "system@*",
  "test@*",           "tutor@*",
  "usenet@*",         "uucp@*"
};
#else
extern char addrstxt[ ATXTS ][ 40 ];
#endif


/* This is the type used for file positioning for fseek, ftell, etc.
/* It can usually be found in sys/types.h as off_t.  If so, the default
/* here should work.  If not, define it yourself, as in the commented out
/* example.  It's usually a long.
*/
#include <sys/types.h>
typedef off_t postype;
/* typedef long postype; */

/* The parameters to use for fopen() calls for read, write, append.  Some
/* systems don't like the "t" for text mode.  If that is the case, delete
/* those and try again.
*/
#define UV_OPEN_READ "rt"
#define UV_OPEN_WRITE "wt"
#define UV_OPEN_APPEND "at"


/*-----------------------------------------------------------------------*/
/* Okay, anything below here is for my benefit.  Do not touch under
/* penalty of death or bungled vote.
*/

#define GROUPS u_groups

/* single item numeric configuration items */
extern int
    u_votetype,                   /* CFV, STV, or STVCFV */
    u_stv,                        /* STV type of vote? */
    u_groups,                     /* number of groups */
    u_needname,                   /* need voter name? */
    u_wantname,                   /* want voter name? */
    u_explicitvote,               /* explicit "vote"? */
    u_maxvotes,                   /* maximum number of votes */
    u_memsort,                    /* do sort in memory? */
    u_dovoteack,                  /* do vote acknowledgement? */
    u_rulesmax,                   /* maximum number of rules */
    u_moderated[MAXGROUPS];       /* group moderated or not */

/* character items */
extern char *u_votename,          /* name of vote */
       *u_voteacksub,             /* acknowledgement subject */
       *u_votemark,               /* ballot marker */
       *u_nametext,               /* name entry text */
       *u_acktxtfile,             /* file for acknowledgement text */
       *u_cmdfile,                /* file for domail script */
       *u_rulesfile,              /* file for rules */
       *groupstxt[MAXGROUPS];     /* group names */


void dump_config();               /* dump usevote.cfg configuration */
void init_global_config();        /* initialize from usevote.cfg */
int space( /* char */ );                /* check if character is whitespace */
int isletter( /* char */ );             /* check if character is letter */
char upcase( /* char */ );              /* convert lower case to upper case */
int strlook( /* char *, char *, int */ );  /* compare strings, ignore case */
void stripstr( /* char * */);          /* strip whitespace from ends of string */
void format( );                   /* format name/addr on line */

#ifndef SEEK_SET
#define SEEK_SET ((postype) 0)
#endif

typedef unsigned short int byte;
typedef unsigned int word;
typedef unsigned long dword;
#define FALSE 0
#define TRUE 1

/* votes - yes, no abstain */
#define NO_V      0
#define YES_V     1
#define ABSTAIN_V 2
#define CANCEL_V  3

/* vote types */
#define CFV_T     0               /* normal CFV */
#define STV_T     1               /* simple STV */
#define STVCFV_T  2               /* STVCFV hybrid */

#endif
