/*---------------------------------------------------------------
/* Produce Vote Ack Lists  (c)1993,94 Ron Dippold All Rights Reserved
/*-------------------------------------------------------------*/
#define VERSION "\nUseVote 3.0 Vote Acknowledge - (c) 1993,94 Ron Dippold\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usevote.h"

#define VRIGHT (RIGHT-GROUPS)   /* address right margin with votes */
FILE *fin=NULL;

char line[200];                   /* input line */
char name[80], addr[80];          /* name and address */
char votestxt[80];
int count, linenum;
int showvote, dosort, altname, pass, erronly;

typedef struct {
  postype pos;                    /* position of vote in file */
  byte vote;                      /* Yes/no vote - for single issue votes */
  char *addr;                     /* pointer to A: info - used for MEMSORT */
} result_type;

result_type *item;

int  steps[6]={51,19,7,3,1,0};     /* for sorting */

char rtag[6] = "ANDCV";
char rname[5][8] = { "Address", "Name", "Date", "Comment", "Vote" };


#define MSG(x) fprintf( stderr, x )


/* -----------------------------------------------------------------
/* Help
/* -----------------------------------------------------------------
*/
void help()
{
  MSG( " uvack (-i) (-n) (-u) (-v) <results>\n\n" );
  MSG("  The -i option only lists the invalid votes section of the ack.\n");
  MSG("  The -n option uses an alternate formatting of names instead of the\n");
  MSG("    regular right-justified format.\n");
  MSG("  The -u option leaves the addresses unsorted.  The sort algorithm\n");
  MSG("    is designed for small memory, not for speed.  You can use your own\n");
  MSG("    faster sorter if you have one.\n");
  MSG( "  The -v option will produce a final vote list which gives all the,\n");
  MSG( "   the yes/no votes.  Don't use this until the final vote result.\n\n");
  exit( 1 );
}

/* ------------------------------------------------------------------
/* exit cleanly
/* ------------------------------------------------------------------
*/
void do_exit( code )
int code;                         /* code to exit with */
{
  int i;

  if( fin != NULL ) fclose( fin );
  if( u_memsort ) {
    for( i=0; i<u_maxvotes; i++ ) {
      if( item[i].addr ) free( item[i].addr ); /* free memory */
    }
  }
  exit( code );
}

/* -----------------------------------------------------------------
/* Strip the '\n' from line
/* -----------------------------------------------------------------
*/
void crstrip()
{
  int j;
  for(j=0; line[j]; j++) {
    if(line[j] == '\n') {
      line[j]=0;
      return;
    }
  }
}


/* -----------------------------------------------------------------
/* Compare string c1 with c2, ignore case.
/* c1<c2 = -1,  c1==c2 = 0,  c1>c2 = 1
/* -----------------------------------------------------------------
*/
int compare( c1, c2 )
char *c1;
char *c2;
{
  for( ;; ) {
    while( *c1 == '"') c1++;
    while( *c2 == '"') c2++;
    if( !*c1 ) {
      if( *c2 ) return( -1 );
                 else return( 0 );
    }
    if( !*c2 ) {
      if( *c1 ) return( 1 );
                 else return( 0 );
    }
    if( ( *c1 | 0x20 ) > ( *c2 | 0x20 ) )
      return( 1 );
    if( ( *c1 | 0x20 ) < ( *c2 | 0x20 ) )
      return( -1 );
    c1++; c2++;
  }
}

/* -----------------------------------------------------------
/* qsort callback routines; these do little more than call compare()
/* -----------------------------------------------------------
*/
int qsort_callback( r1 , r2 )
result_type *r1;
result_type *r2;
{
  static int i=0, j=3;
  static char spin[]="/-\\|";
  if( !( ++i % 100 ) ) {
    putc( spin[(++j)%4], stderr );  putc( '\b', stderr );
  }

/*
  printf( "A1 %3d: %s\nA2 %3d: %s\n%d\n",
    r1-&item[0],  r1->addr,
    r2-&item[0],  r2->addr,
    compare( r1->addr, r2->addr));
  getch();
*/

  return( compare( r1->addr, r2->addr ));
}

/* ----------------------------------------------------
/* Do the sort
/* ----------------------------------------------------
*/
void sortaddr()
{

  if( u_memsort ) {               /* memory sort */
    int i, j;
    char * buf_pointer;

    fprintf( stderr, "Memory Quicksorting %d votes... ", count);
    qsort(item, count, sizeof(result_type), qsort_callback);
    MSG(" \n\n");

  } else {                        /* no memory sort */

    int  pass, step, swapped, i, k, i1, j, which, start, end;
    result_type titem;
    static char cmp[2][80];

    fprintf( stderr, "Disk Shellsorting %d votes...", count);
    pass = 0;

    do {
      if ( !(step = steps[pass]) ) {
        MSG("\n\n");
        return;                        /* all done */
      }
      pass++;
      fprintf( stderr, "\n  Pass %d: [", pass);
      for( i=0; i<step; i++ ) fputc( ' ', stderr );
      fputc( ']', stderr );
      for( i=0; i<=step; i++ ) fputc( '\b', stderr );

      for( start=0; start<step; start++) {
        end = count;
        MSG( ".");
        swapped = 1;
        while( start+step < end && swapped ) {
          i1 = start;
          which=0;
          fseek( fin, item[i1].pos, SEEK_SET );        /* read quote */
          fgets( cmp[which], 80, fin );
          swapped = 0;

          for( j=start+step; j<end; j+=step) {
            fseek( fin, item[j].pos, SEEK_SET );        /* read quote */
            fgets( cmp[which^1], 80, fin );

            k = compare( cmp[which], cmp[which^1]);
            if ( k > 0) {   /* swap i and j */
              titem = item[i1]; item[i1] = item[j]; item[j] = titem;
              swapped = 1;
            } else {
              which = which^1;                 /* other buffer */
            }
            i1 = j;

          }  /* for j= */
          end -= step;
        }  /* do while */
      } /* for start */
    }  while (1); /* do */
    MSG("\n\n");
  }
}

/* ----------------------------------------------------
/* Print the names of all the groups for a vote list
/* ----------------------------------------------------
*/
void printgroups()
{
  int i, j;

  printf( "\n\n" );
  for( i=GROUPS-1; i>=0; i-- ) {
    j = RIGHT-GROUPS-4-strlen( groupstxt[i] );
    if( j<0 ) j=0;
    while( j-- ) printf( " " );
    printf( "%s ", groupstxt[i] );
    for( j=0; j<i+3; j++ ) printf( "-" );
    printf( "+" );
    for( j=i; j<GROUPS-1; j++ ) printf( "|" );
    printf( "\n" );
  }
  for( j=0; j<RIGHT-GROUPS; j++ ) printf( " " );
  for( j=0; j<GROUPS; j++ ) printf( "|" );
  printf( "\n" );

}



/* ----------------------------------------------------
/* Main procedure - almost everything goes here
/* ----------------------------------------------------
*/
int main( argc, argv )
int argc;
char **argv;
{
  int i, j, vabstain, verror;
  long mtot;

  init_global_config();
  item = (result_type*)malloc(u_maxvotes*sizeof(result_type));

  MSG( VERSION );
  MSG( "  Don't forget to 'uvdup' to get rid of duplicates first!\n");

  i = 1;
  altname = FALSE;
  showvote = FALSE;
  dosort = TRUE;
  erronly = FALSE;

  while( i<argc && argv[i][0] == '-' ) {
    if( argv[i][1] == 'v' ) {
      showvote = TRUE;
    } else if( argv[i][1] == 'u' ) {
      dosort = FALSE;
    } else if( argv[i][1] == 'n' ) {
      altname = TRUE;
    } else if( argv[i][1] == 'i' ) {
      erronly = TRUE;
    } else {
      help();
    }
    i++;
  }
  if( i!=argc-1 ) help();

  if( u_memsort ) {
    for( j=0; j<u_maxvotes; item[j++].addr = NULL ); /* null out pointers */
  }

  if ( !( fin = fopen( argv[i], UV_OPEN_READ ))) {
    fprintf( stderr, "Can't open %s for input\n", argv[i] );
    do_exit( 2 );
  }

  printf("%s", u_votename );
  if( showvote ) {
    printf(" Final Vote Ack\n");
  } else {
    printf(" Mass Ack\n");
  }

  vabstain = verror = FALSE;

  /* Read all the votes for sorting */
  count = linenum = 0;
  for( ;; ) {
    if( count >= u_maxvotes ) {
      fprintf(stderr, "  ERROR!  Exceeded maximum of %d votes.  Check usevote.cfg\n",u_maxvotes);
      do_exit(30);
    }
    item[count].pos = ftell( fin );
    if( !fgets( line, 200, fin ) ) {
      break;                      /* exit from loop */
    }
    if( u_memsort ) {
      for( j=0; line[j]!='\n'; j++ ); /* no lf... */
      line[j++]=0;
      mtot += j;
      if( (item[count].addr = ( char * ) malloc( j )) == NULL ) {
        fprintf( stderr, "  Not enough memory for qsort.  %ld\n%s",mtot,
          "  Try recompiling with MEMSORT disabled in usevote.h.\n");
        do_exit(99);
      } else {
        memcpy( item[count].addr, &line[3], j-3 );
      }
    }
    linenum++;
    for( j=0; j<5; j++ ) {
      if( j > 0 ) {
        linenum++;
        if( !fgets( line, 200, fin ) ) {
          printf("Error!  Unexpected end of file line %d!\n", linenum);
          do_exit(10+j);
        }
      }
      if( line[0]!=rtag[j]) {
        printf("Error!  %s not found line %d \n", rname[j], linenum);
        do_exit(10+j);
      }
    }
    item[count].vote = line[3];
    if( item[count].vote == '-' ) vabstain = TRUE; /* votes abstained */
    if( item[count].vote == '!' ) verror = TRUE;   /* vote in error */
    count++;
  }

  if( count == 0 ) {
    MSG("  ERROR!  No votes!\n");
    do_exit(99);
  }

  if( dosort ) {
    sortaddr();
  }

  pass = 0;
  if( !erronly ) {
    for ( ;; ) {
      if( GROUPS == 1 && showvote ) {
        if( pass == 0 ) {
          printf("\nVoted Yes\n");
        } else if ( pass == 1 ) {
          printf("\nVoted No\n");
        } else {
          printf("\nAbstained\n");
        }
      }
      if( GROUPS == 1 || !showvote ) {
        for(i=0; i<RIGHT; line[i++]='-');
        line[i] = 0;
        puts(line);
      }

      if( showvote &&  GROUPS>1 ) {
        printgroups();
      }

      for( i=0; i<count; i++) {

        /* don't show error votes this pass */
        if( item[i].vote == '!' ) continue;

        /* never show canceled votes */
        if( item[i].vote == '*' ) continue;

        /* show only the right votes on the right pass */
        if( GROUPS == 1 && showvote ) {
          if( pass == 0 && item[i].vote != 'Y' ) continue;
          if( pass == 1 && item[i].vote != 'N' ) continue;
          if( pass == 2 && item[i].vote != '-' ) continue;
        }

        fseek( fin, item[i].pos, SEEK_SET ); /* read vote */
        fgets( line, 80, fin );   /* get address */
        crstrip();
        strcpy( addr, &line[3] );

        fgets( line, 80, fin );                /* get name */
        crstrip();
        strcpy( name, &line[3] );

        fgets( line, 80, fin );                /* get votes */
        fgets( line, 80, fin );                /* get votes */
        fgets( line, 80, fin );                /* get votes */
        if( showvote == 1 ) {
          crstrip();
          strcpy( votestxt, &line[2] );
        }

        if( showvote == 1 && GROUPS>1 ) {
          format( line, addr, name, altname, VRIGHT-1 );
          strcat( line, votestxt );
        } else {
          format( line, addr, name, altname, RIGHT );
        }
        printf("%s\n",line);
      }  /* for */
      pass++;
      if( GROUPS != 1 || !showvote ) break; /* break out */
      if( pass == 2 && !vabstain ) break;  /* no abstains */
      if( pass == 3 ) break;        /* no matter what */
    }
  }

  if( verror ) {                  /* errors! */
    printf( "\n\nVotes in error\n" );
    for(i=0; i<RIGHT; line[i++]='-');
    line[i] = 0;
    puts(line);

    for( i=0; i<count; i++) {

      /* only show error votes */
      if( item[i].vote != '!' ) continue;

      fseek( fin, item[i].pos, SEEK_SET ); /* read vote */
      fgets( line, 80, fin );     /* get address */
      crstrip();
      strcpy( addr, &line[3] );

      fgets( line, 80, fin );     /* get name */
      crstrip();
      strcpy( name, &line[3] );

      fgets( line, 80, fin );
      fgets( line, 80, fin );
      fgets( line, 80, fin );     /* get votes */
      crstrip();
      strcpy( votestxt, &line[2] );
      votestxt[RIGHT]=0;          /* right margin */

      format( line, addr, name, altname, RIGHT );
      printf("%s\n",line);
      printf("  %s\n", votestxt );/* why rejected */
    }  /* for */
  }
  do_exit( 0 );
}

