/*---------------------------------------------------------------
/* Produce Bounce Ack Lists  (c)1993,94 Ron Dippold All Rights Reserved
/*-------------------------------------------------------------*/
#define VERSION "\nUseVote 3.0 Bounce List - (c) 1993,94 Ron Dippold\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usevote.h"

FILE *fin;

char line[200];                   /* input line */
postype *pos;                     /* position of vote in file */

int altname;

int dosort, count;
int  steps[5]={19,7,3,1,0};     /* for sorting */
#define MSG(x) fprintf( stderr, x )

/* -----------------------------------------------------------------
/* Help
/* -----------------------------------------------------------------
*/
void help()
{
  MSG( "\n  uvbounce (-n) (-u) <bounce file>\n\n" );
  MSG( "  <bounce file> contains the raw contents of the bounced\n" );
  MSG( "  Ack messages that were sent returned.  It looks for the\n" );
  MSG( "  Voter: lines in the messages, then sorts and prints them.\n" );
  MSG( "  The -n option uses an altername name formatting method.\n" );
  MSG( "  The -u option will prevent the list from being sorted.\n" );
  exit( 1 );
}

/* -----------------------------------------------------------------
/* Compare string c1 with c2, ignore case, return 1 if c1>c2
/* -----------------------------------------------------------------
*/
int compare( c1, c2 )
char *c1;
char *c2;
{
  do {
    if( ( *c1 | 0x20 ) > ( *c2 | 0x20 ) )
      return( 1 );
    if( ( *c1 | 0x20 ) < ( *c2 | 0x20 ) )
      return( 0 );
  } while ( *c1++ && *c2++ );
  return( 0 );
}


/* ----------------------------------------------------
/* Do the sort - it's a shellsort
/* ----------------------------------------------------
*/
void sortaddr()
{
  int  pass, step, swapped, i, k, i1, j, which, start, end;
  postype tpos;
  static char cmp[2][100];

  fprintf( stderr, "Sorting %d addresses...", count);
  pass = 0;

  do {
    if ( !(step = steps[pass]) ) {
      MSG("\n\n");
      return;                        /* all done */
    }
    pass++;
    fprintf( stderr, "\n  Pass %d: [", pass);
    for( i=0; i<step; i++ ) fputc( ' ', stderr );
    fputc( ']', stderr );
    for( i=0; i<=step; i++ ) fputc( '\b', stderr );

    for( start=0; start<step; start++) {
      end = count;
      MSG( ".");
      swapped = 1;
      while( start+step < end && swapped ) {
        i1 = start;
        which=0;
        fseek( fin, pos[i1], SEEK_SET );
        fgets( cmp[which], 80, fin );
        swapped = 0;

        for( j=start+step; j<end; j+=step) {
          fseek( fin, pos[j], SEEK_SET );        /* read quote */
          fgets( cmp[which^1], 80, fin );

          k = compare( cmp[which], cmp[which^1]);
          if ( k > 0) {   /* swap i and j */
            tpos = pos[i1]; pos[i1] = pos[j]; pos[j] = tpos;
            swapped = 1;
          } else {
            which = which^1;                 /* other buffer */
          }
          i1 = j;

        }  /* for j= */
        end -= step;
      }  /* do while */
    } /* for start */
  }  while (1); /* do */
  MSG("\n\n");
}


/* ----------------------------------------------------
/* Main procedure - almost everything goes here
/* ----------------------------------------------------
*/
int main( argc, argv )
int argc;
char **argv;
{
  int i;
  static char last[200], addr[200], name[200];

  init_global_config();
  pos = (postype*) malloc(u_maxvotes*sizeof(postype));

  MSG( VERSION );

  i = 1;
  dosort = TRUE;
  altname = FALSE;

  while( i<argc && argv[i][0] == '-' ) {
    if( argv[i][1] == 'u' ) {
      dosort = FALSE;
    } else if( argv[i][1] == 'n' ) {
      altname = TRUE;
    } else {
      help();
    }
    i++;
  }
  if( i!=argc-1 ) help();

  if ( !( fin = fopen( argv[i], UV_OPEN_READ ))) {
    fprintf( stderr, "Can't open %s for input\n", argv[i] );
    exit( 2 );
  }


  /* Read all the votes for sorting */
  count = 0;
  for( ;; ) {
    if( count >= u_maxvotes ) {
      fprintf(stderr, "  ERROR!  Exceeded maximum of %d votes.  Check usevote.cfg\n",u_maxvotes);
      exit(30);
    }
    pos[count] = ftell( fin );
    if( !fgets( line, 200, fin ) ) {
      break;                      /* exit from loop */
    }
    if( strlook( line, "Voter address: ", 15 ) == 0 ) {
      count++;
    }
  }

  if( dosort && count!=0 ) {
    sortaddr();
  }

  printf("%s Bounce List - No need to revote\n",u_votename);
  for(i=0; i<RIGHT; line[i++]='-');
  line[i] = 0;
  puts(line);

  if( count == 0 ) {
    printf( "No bounced votes at this time\n" );
  }

  strcpy( last, "" );
  for( i=0; i<count; i++ ) {
      fseek( fin, pos[i], SEEK_SET ); /* read vote */
      fgets( line, 100, fin );    /* get address */
      strcpy( addr, &line[15] );
      stripstr( addr );
      fgets( line, 100, fin );    /* get name */
      if( strlook( line, "Voter name:", 11 ) == 0 ) {
        strcpy( name, &line[15] );
        stripstr( name );
      } else {
        strcpy( name, "" );
      }
      if( strcmp( last, addr ) != 0 ) {
        format( line, addr, name, altname, RIGHT );
        printf( "%s\n", line );
      }
      strcpy( last, addr );
  }

  fclose(fin);
  return( 0 );
}

