/*---------------------------------------------------------------
/* Produce CFV and Result  (c)1993 Ron Dippold All Rights Reserved
/*-------------------------------------------------------------*/
#define VERSION "\nUseVote 3.0 Simple CFV Processor - (c) 1993 Ron Dippold\n\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usevote.h"

#define LINESIZE 8000
char line[LINESIZE];              /* input line */
char line2[LINESIZE];             /* input line */
FILE *fin=NULL;

char gcomtxt[MAXGROUPS][80];      /* group comment text */

int groot;                        /* group root text position if any */
char groot_txt[80];               /* root text */

char mroot_txt[80];               /* message id root */

/* Define the different documents we can produce */
#define RFD 0
#define CFV1 1
#define CFV2 2
#define RESULT 3
#define WLAST  4
int which;

char sub_txt[WLAST][10] = { "RFD", "CFV", "2nd CFV", "RESULT" };
char mid_txt[WLAST][10] = { "RFD1", "CFV1", "CFV2", "RESULT" };
int mid_len[WLAST];

#define H_FROM 0
#define H_REPLY_TO 1
#define H_FOLLOWUP_TO 2
#define H_NEWSGROUPS 3
#define H_SUBJECT 4
#define H_MESSAGE_ID 5
#define H_REFERENCES 6
#define H_SUPERSEDES 7
#define H_ORGANIZATION 8
#define H_APPROVED 9
#define H_EXPIRES 10
#define H_COUNT 11
char *head[H_COUNT];              /* store text of headers */
int inheader;                     /* if we're in the header */

char h_key[H_COUNT][20]= {        /* header keys to process */
  "From: ",
  "Reply-To: ",
  "Followup-To: ",
  "Newsgroups: ",
  "Subject: ",
  "Message-ID: ",
  "References: ",
  "Supersedes: ",
  "Organization: ",
  "Approved: ",
  "Expires: "
  };
int h_klen[H_COUNT];              /* length of header keys */

#define I_COUNT 6
char i_key[I_COUNT][25] = {            /* headers to ignore */
  "Path: ",
  "Date: ",
  "Lines: ",
  "Sender: ",
  "NNTP-Posting-Host: ",
  "Xref: "
};
int i_klen[I_COUNT];

char title_txt[WLAST][40] = {
  "",                             /* no need for RFD */
  "FIRST CALL FOR VOTES (of 2)",
  "LAST CALL FOR VOTES (of 2)",
  "RESULT"
};
int title_len[WLAST];




/* --------------------------------------------------------------------
/* Bail out!
/* --------------------------------------------------------------------
*/
void do_exit( msg, level )
char *msg;                        /* message to print */
int level;                        /* error level */
{
  int i;

  if( fin ) fclose( fin );
  for( i=0; i<H_COUNT; i++ )
    if( head[i] ) free( head[i] );
  exit( level );
}


/* --------------------------------------------------------------------
/* Process a header line
/* --------------------------------------------------------------------
*/
void header_process( )
{
  int i, j;

  j = FALSE;
  for( i=0; i<H_COUNT; i++ ) {
    if( strlook( line, h_key[i], h_klen[i] ) == 0 ) { /* matches header */
      j = TRUE;
      if( !head[i] ) {     /* should be unused */
        if( !( head[i] = malloc( strlen( line )))) {
          do_exit( "\n  ERR: Couldn't allocate memory for header.\n", 5 );
        }
        strcpy( head[i], &line[ h_klen[i] ] );
      } else {
        fprintf( stderr, "  WARNING: Duplicate header '%s' ignored\n", h_key[i] );
      }
    }
  }

  if ( !j ) {                     /* headers to just ignore */
    for( i=0; i<I_COUNT; i++ )
      if( strlook( line, i_key[i], i_klen[i] ) == 0 ) j = TRUE;
  }

  if ( !j ) {                     /* no match */
    for( i=0; line[i]!=' ' && line[i]; i++ ); /* terminate */
    line[i] = 0;
    fprintf( stderr, "  Header '%s' discarded.\n", line );
  }
}


/* --------------------------------------------------------------------
/* Sort groups and print them out, eliminating news.announce.newgroups
/* and news.groups, which are implied
/* --------------------------------------------------------------------
*/
void sort_groups( list, dosort )
char  *list;                      /* list to sort */
int   dosort;                     /* true if really sort */
{
#define MAX 50                    /* 50 groups!  Overkill. */
  static int idx[MAX];
  static int ilen[MAX];           /* length of name */
  int i, j, k, t, cnt;

  i=j=0;
  while ( list[j] ) {             /* repeat */
    idx[i] = j;                   /* start of group name */
    while( list[j]!=',' && list[j]!=' ' && list[j] ) j++; /* end of name */
    ilen[i] = j-idx[i];           /* length of name */
    if( strlook( "news.announce.newgroups", &list[idx[i]], 23 ) == 0 ||
        strlook( "news.groups", &list[idx[i]], 11 ) == 0 ) i--;
    i++;                          /* next one */
    while( list[j]==',' || list[j]==' ' ) j++; /* skip over */
    if( i == MAX ) do_exit( "  ERR: 50 groups?  You're insane!\n", 6 );
  }
  cnt = i;

  for( i=0; i<cnt; i++ ) {
    list[idx[i]+ilen[i]] = 0;     /* terminate group names */
  }


  if( cnt>1 && dosort ) {         /* do the sort */
    for( i=cnt-1; i>0; i-- ) {      /* so it's a bubblesort... nyah! */
      for( j=0; j<i; j++ ) {
        for( k=0; k<ilen[j] || k<ilen[j+1]; k++ ) {
          if( list[idx[j]+k] > list[idx[j+1]+k] ) {
            t = idx[j+1]; idx[j+1] = idx[j]; idx[j] = t;
            t = ilen[j+1]; ilen[j+1] = ilen[j]; ilen[j] = t;
            break;
          }
          if( list[idx[j]+k] < list[idx[j+1]+k] ) break;
        }                         /* for k */
      }                           /* for j */
    }                             /* for i */
  }

  for( i=0; i<cnt; i++ ) {
    printf( ",%s", &list[idx[i]] );  /* print group name */
  }
  printf( "\n" );
}


/* --------------------------------------------------------------------
/* From mroot_txt ( message ID root ), print a message ID
/* --------------------------------------------------------------------
*/
void printid( which )
int which;
{
  int i;

  for( i=0; mroot_txt[i]!='#'; i++ ) /* find "#" in mroot */
    printf( "%c", mroot_txt[i] );
  printf( "%s", mid_txt[which] ); /* print which one */
  while( mroot_txt[++i] )           /* print the rest */
    printf( "%c", mroot_txt[i] );
/*  printf( " " ); */
}


/* --------------------------------------------------------------------
/* Finish off the header - tricky!
/* --------------------------------------------------------------------
*/
void header_finish( )
{
  int h, i, j, k, g, i1;

  /* From: */
  h = H_FROM;
  printf( "%s", h_key[h] );
  if( head[h] ) {
    printf( "%s\n", head[h] ); /* print as is */
  } else {
    printf( "[!!! Your email address]\n" );
    fprintf( stderr, "  WARNING: From needs help\n" );
  }

  /* Reply-To: */
  h++;
  if( which == RESULT ) {
    ;                             /* not in Result! */
  } else {
    printf( "%s", h_key[h] );
    if( head[h] ) {
      printf( "%s\n", head[h] ); /* print as is */
    } else {
      printf( "[!!! Your voting mail address]\n" );
      fprintf( stderr, "  WARNING: Reply-To needs help\n" );
    }
  }

  /* Followup-To: */
  h++;
  printf( "%s", h_key[h] );
  if( which == RESULT ) {
    printf( "news.groups\n" );    /* always followup to news.groups */
  } else {
    printf( "poster\n" );
  }

  /* Newsgroups */
  h++;
  printf( "%snews.announce.newgroups,news.groups", h_key[h] );
  if( head[h] ) {
    sort_groups( head[h], which == CFV1 ); /* sort on 1st CFV */
  } else {
    printf( ",[!!! groups]\n" );
    fprintf( stderr, "  WARNING: Newsgroups needs help\n" );
  }

  /* Subject: */
  h++;
  printf( "%s%s", h_key[h], sub_txt[which] );
  j = FALSE;
  if( head[h] ) {
    for( i=0; head[h][i]!=':' && head[h][i]; i++ );
    if( head[h][i] ) j = TRUE;
  }
  if( j ) {                       /* header already exists */
    printf( "%s", &head[h][i] );  /* just print what's there */
  } else {                        /* construct header */
    printf( ": " );
    if ( GROUPS == 1 ) {          /* one group is easy */
      printf( "%s", groupstxt[0] );
    } else {
      if( groot !=-1 ) {          /* yes we can! */
        printf( "%s.{", groot_txt );
        for( g=0; g<GROUPS; g++ )
          printf( "%s%c", &groupstxt[g][groot], ( g==GROUPS-1 ) ? '}' : ',' );
      } else {
        printf( "[!!! vote description]" );
        fprintf( stderr, "  WARNING: Subject needs help\n" );
      }

    }
  }
  if( which == RESULT ) printf( " [!!! result]" );
  printf( "\n" );                 /* finish off header */


  /* Message-ID: */
  h++;
  printf( "%s", h_key[h] );
  j = -1;
  if( head[h] ) {                 /* already have a header */
    for( i=0; i<strlen( head[h] ); i++ )  /* check every position */
      for( i1=0; i1<WLAST; i1++ )  /* check for every string */
        if( strlook( &head[h][i], mid_txt[i1], mid_len[i1] ) == 0 ) {
          j = i1;
          k = i;
          break;
        }
    if( j != -1 ) {               /* preserve old ID */
      head[h][k] = 0;
      sprintf( mroot_txt, "%s#%s", head[h], &head[h][k+mid_len[j]] );
    }
  }
  if( j == -1 ) {                 /* build new message-ID */
    if( GROUPS == 1 ) {
      sprintf( mroot_txt, "<%s", groupstxt[0] );
    } else {
      if( groot != -1 ) {
        sprintf( mroot_txt, "<%s-reorg", groot_txt );
      } else {
        sprintf( mroot_txt, "<[!!! vote summary]" );
        fprintf( stderr, "  WARNING: Message-ID needs help\n" );
      }
    }
    strcat( mroot_txt, "-#@uunet.uu.net>" );
  }
  printid( which );
  printf( "\n" );


  /* References: */
  h++;
  printf( "%s", h_key[h] );
  g=FALSE;                        /* assume no references */
  if( head[h] ) {                 /* already existing header */
    i=0;
    for( ;; ) {
      while( head[h][i] && head[h][i]!='<' ) i++; /* start of message ID */
      if( !head[h][i] ) break;    /* end */
      j=TRUE;                     /* assume it will be based on root */
      for( k=0; mroot_txt[k]!='#'; k++ ) {
        if( mroot_txt[k] != head[h][i+k] ) {
          j = FALSE;              /* no match */
          break;
        }
      }
      if( !j ) {                  /* no match, print it */
        while( head[h][i]!='>' ) printf( "%c", head[h][i++] );
        printf( "> " );
        g = TRUE;
      }
      i++;
    }                             /* do - must break out of */
  }
  if( !g ) printid( RFD );    /* only RFD if no references */
  if( which>=CFV2 ) printid( CFV1 );
  if( which>=RESULT ) printid( CFV2 );
  printf( "\n" );

  /* Supersedes: */
  h++;
  if( which != CFV1 ) {
    printf( "%s", h_key[h] );
    printid( which - 1 );
    printf( "\n" );
  }

  /* Organization */
  h++;
  printf( "%s", h_key[h] );
  if( head[h] ) {
    printf( "%s\n", head[h] ); /* print as is */
  } else {
    printf( "Usenet Volunteer Votetakers\n" );
  }

  /* Approved */
  h++;
  printf( "%s", h_key[h] );
  if( head[h] ) {
    printf( "%s\n", head[h] ); /* print as is */
  } else {
    printf( "tale@uunet.uu.net\n" );
  }


  /* Expires */
  h++;
  if( which != RESULT ) {
    printf( "%s", h_key[h] );
    if( head[h] ) {
      printf( "%s\n", head[h] ); /* print as is */
    } else {
      printf( " [!!! 22 days]\n" );
      fprintf( stderr, "  WARNING: Expires needs help\n" );
    }
  }
}


/* --------------------------------------------------------------------
/* Process a line of body
/* --------------------------------------------------------------------
*/
void body_process( )
{
  int i, j, k, l;
  static char vm1[40], vm2[40];

  /* convert tabs to spaces - 8 stop */
  for( i=k=0; line[i]; i++ ) {
    if( line[i] == '\t' ) {
      do {
        line2[k++] = ' ';
      } while ( ( k%8 ) > 0 );
    } else {
      line2[k++] = line[i];
    }
  }
  line2[k++]=0;
  strcpy( line, line2 );

  /* look for post header line */
  for( k=i=0; line[i]==' '; i++ );
  for( j=1; j<WLAST; j++ ) {
    if( strlook( &line[i], title_txt[j], title_len[j] ) == 0 ) {
      k=j;
    }
  }
  if( k!=0 ) {                    /* it was */
    j = i*2 + title_len[k];       /* find right margin */
    j = ( j-title_len[which] )/2; /* find left position */
    for( i=0; i<j; i++ ) printf( " " );
    printf( "%s\n", title_txt[which] );
    return;
  }

  /* look for a vote mark */
  if( u_votemark ) {
    for( j=0; u_votemark[j]; j++ ) {
      if( u_votemark[j] == '%' ) {
        vm1[j] = '1';
        vm2[j] = '2';
      } else {
        vm1[j] = vm2[j] = u_votemark[j];
      }
    }
    vm1[j] = vm2[j] = 0;
    if( which == CFV2 || which == CFV1 ) {
      for( i=0; line[i]; i++ ) {
        if( strlook( &line[i], vm1, j ) == 0 ||
            strlook( &line[i], vm2, j ) == 0 ) {
          if( which == CFV1 ) memcpy( &line[i], vm1, j );
          if( which == CFV2 ) memcpy( &line[i], vm2, j );
          break;
    } } }
  }

  /* replace [group] with group name and [moderated] with status for
  /* 1 group vote
  */
  if( GROUPS == 1 ) {
    for( i=0; line[i]; i++ ) {
      if( strlook( &line[i], "[group]", 7 ) == 0 ) {
        line[i]=0;                /* terminate */
        sprintf(line2, "%s%s%s", line, groupstxt[0], &line[i+7] );
        strcpy( line, line2 );
      }
    }
    for( i=0; line[i]; i++ ) {
      if( strlook( &line[i], "[moderated]", 11 ) == 0 ) {
        line[i]=0;                /* terminate */
        sprintf(line2, "%s%s%s", line, u_moderated[0] ? "" : "un", &line[i+11] );
        strcpy( line, line2 );
      }
    }
  }

  /* center lines */
  if( strlook( line, "[center]", 8 ) == 0 ) { /* center line */
    strcpy( line2, &line[8] );
    stripstr( line2 );
    i = ( RIGHT-strlen(line2))/2;
    line[i] = 0;                  /* end it */
    while( --i>=0 ) line[i] = ' ';
    strcat( line, line2 );
    strcat( line, "\n" );
  }

  /* centered group list */
  if( strlook( line, "[grouplist]", 11 ) == 0 ) {
    for( i=0; i<GROUPS; i++ ) {
      sprintf( line2, "%smoderated group %s%s", u_moderated[i] ? "" : "un",
        groupstxt[i], gcomtxt[i] );
      j = ( RIGHT-strlen(line2))/2;
      line[j] = 0;                  /* end it */
      while( --j>=0 ) line[j] = ' ';
      strcat( line, line2 );
      strcat( line, "\n" );
      if( i<GROUPS-1 ) printf( "%s", line );
    }
  }

  printf( "%s", line );
}


/* --------------------------------------------------------------------
/* Print help
/* --------------------------------------------------------------------
*/
void help()
{
  printf( "  uvcfv <which> <cfv file>\n\n" );
  printf( "   <which> should be 1, 2, or r for 1st or 2nd CFV or Result.\n\n" );
  printf( "   This program helps massage your CFVs into shape for mailing\n" );
  printf( "   to tale@uunet.uu.net.  It will try to fill in as many headers\n" );
  printf( "   as possible and flag the rest.  It does some simple fill in the\n" );
  printf( "   blank processing.  Check the result carefully!\n\n" );
  exit(2);
}


/* --------------------------------------------------------------------
/* Main procedure - This does a lot of stuff.
/* --------------------------------------------------------------------
*/

int main( argc, argv )
int argc;
char **argv;
{

  int i, j, g;

  fprintf( stderr, VERSION );
  if( argc!=3 ) help();

  switch ( argv[1][0] ) {
    case '1':
      which = CFV1;
      break;
    case '2':
      which = CFV2;
      break;
    case 'R':
    case 'r':
      which  = RESULT;
      break;
    default:
      help();
  }

  init_global_config();
  for( i=0; i<GROUPS; i++ ) {     /* no comments in group names */
    strcpy( gcomtxt[i], "" );
    for( j=0; groupstxt[i][j]; j++ ) {
      if( space( groupstxt[i][j] )) {
        strcpy( gcomtxt[i], &groupstxt[i][j] );
        groupstxt[i][j] = '\0';
        break;
      }
    }
  }

  inheader = TRUE;               /* not in header yet */
  for( i=0; i<H_COUNT; i++ ) {    /* set up for headers */
    head[i] = NULL;
    h_klen[i] = strlen( h_key[i] );
  }
  for( i=0; i<I_COUNT; i++ ) i_klen[i] = strlen( i_key[i] );
  for( i=0; i<WLAST; i++ ) mid_len[i] = strlen( mid_txt[i] );
  for( i=0; i<WLAST; i++ ) title_len[i] = strlen( title_txt[i] );

  /* check the group names to see if we can foo.bar.{bletch,barf} */
  groot = -1;
  if( GROUPS != 1 ) {
    for( i=0; groupstxt[0][i]; i++ ); /* find end of group name */
    while( groupstxt[0][--i] !='.' ); /* find last period */
    groot = i+1;             /* position of last period+1 */
    for( g=1; g<GROUPS; g++ )  /* for each group */
      for( i=0; i<groot; i++ )
        if( groupstxt[g][i] != groupstxt[0][i] || groupstxt[g][i]=='.' ) {
          groot=-1;
          break;
        }
    for( i=0; i<groot-1; i++ )
      groot_txt[i] = groupstxt[0][i];
    groot_txt[i] = '\0';
  }

  if ( !( fin = fopen( argv[2], UV_OPEN_READ ))) {
    printf( "Can't open %s for input\n", argv[2] );
    exit( 2 );
  }

  for( ;; ) {
    if( !fgets( line, LINESIZE, fin ) ) {
      break;                      /* exit from loop */
    }

    /* one header "line" can stretch over several physical lines. Consolidate
    /* them here, since that's the elegant thing to do
    */
    if( inheader ) {
      if( line[0]=='\n' ) {       /* blank line means no more header */
        inheader = FALSE;
        header_finish();          /* finish up the header first */
      } else {
        char c;
        for( i=0;; ) {            /* check for more */
          for( j=0; i<LINESIZE-2 && line[i]!='\n'; i++ ) { /* end of line */
            if( line[i]!=' ' && line[i] != '\t' ) j=1; /* all blank line */
          }
          if( j == 0 ) {          /* all blank line! */
            c = '\n';
            ungetc( c, fin );     /* push all blank line */
          } else {
            c = getc( fin );        /* peek at next line */
            ungetc( c, fin );       /* put it back */
          }
          if( c == ' ' || c == '\t' ) { /* whitespace, read next line */
            line[i] = 0;          /* just in case error... */
            fgets( &line[i], LINESIZE-i-2, fin );
          } else {
            line[i++] = 0;
            break;
          }
        }
      }
    }

    if ( inheader ) {
      header_process();
    } else {
      body_process();
    }
  }                               /* for each line */
  fclose( fin );  fin = NULL;

  /* append bounce.lst if we can */
  if( which == CFV2 ) {
    if ( fin = fopen( "bounce.lst", UV_OPEN_READ )) {
      printf( "\n\n" );
      for( ;; ) {
        if( !fgets( line, LINESIZE, fin ) ) {
          break;                      /* exit from loop */
        }
        printf( "%s", line );
      }
    }
  }

  do_exit( "\nCheck the result before using, and mail to newgroups@uunet.uu.net!\n", 0 );
  return( 0 );                    /* never happens */
}
