/*---------------------------------------------------------------
/* Produce Vote Count  (c)1993,94 Ron Dippold All Rights Reserved
/*-------------------------------------------------------------*/
#define VERSION "\nUseVote 3.0 Vote Count - (c) 1993,94 Ron Dippold\n\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usevote.h"

/* For Non-STV voting */
int *novote,*yesvote,*absvote;
char **passdate;
char date[80];

/* For STV voting */
#define MAX 100
#define NVOTE GROUPS
int total[MAXGROUPS+1];           /* totals for each group */
char **votes;                     /* actual votes */
int active[MAXGROUPS+1];          /* if group is still active */
int idx[MAXGROUPS+1];             /* sort group totals... */
char gtxt[MAXGROUPS+1][80];

FILE *fin;

char line[200];                   /* input line */
int count,cond1,cond2,badcount;
int dodate, dorepeat, dotable;    /* conditionals */

char noyes[2][4]= { " No","Yes" };
char failpass[2][7]= { "fails", "passes" };

char rtag[6] = "ANDCV";
char rname[5][8] = { "Address", "Name", "Date", "Comment", "Vote" };

void help()
{
  printf( "  uvcount (-d) (-r) (-t) <results>\n\n" );
  printf( "  -d (non-STV votes) prints the date field from the vote that put the\n");
  printf( "  vote 'over the top' assuming the group passes.\n\n");
  printf( "  -r (STV votes) forces uvcount to remove the winner and restart to find\n" );
  printf( "  the next most popular choices.  Repeat till only one left.\n\n" );
  printf( "  -t (non-STV votes) forces even single group voting results to use the\n");
  printf( "  multigroup table format.\n");
  exit(2);
}

/* Word wrap output to right margin
*/
void wrapout( txt )
char *txt;
{
  static int x=0;                 /* x position */
  static char build[160];         /* build line */
  int i, j, k;                    /* counters */

  j = 0;
  while( txt[j] ) {
    if( txt[j] == '\n' || txt[j] == ' ' ) {
      if( x >= RIGHT ) {          /* need to wrap */
        for(i=x; build[i]!=' ' || i>=RIGHT; i-- ); /* find space */
        build[i++]=0; /* print it */
        printf( "%s\n", build );
        for( k=0; i<x; build[k++] = build[i++] );  /* copy rest */
        x = k;
      }
      if( txt[j] == '\n' ) {
        build[x]=0;
        printf( "%s\n", build );
        x=0;
      }
    }
    if( txt[j] !='#' ) build[x++] = txt[j++];        /* copy text */
                  else j++;
  }
}

/* --------------------------------------------------------------------
/* Tally STV results
/* --------------------------------------------------------------------
*/
int tally_stv()
{
  int i, j, k;

  for( ;; ) {                   /* count the votes */
    for( i=0; i<=NVOTE; i++ ) total[i]=0;
    for( i=0; i<count; i++ ) {  /* for each vote */
      int found, repeat;
      byte lookfor;

      lookfor = '1';              /* look for 1st choice... */
      for( repeat = TRUE; repeat; ) {
        found = FALSE;            /* no match found this time */
        for( j = 0; j<GROUPS; j++ ) {
          if( votes[i][j] == lookfor ) {
            found = TRUE;
            if( active[j] ) {     /* group active */
              total[j]++;         /* add it! */
              repeat = FALSE;     /* don't repeat */
              break;
            } else {              /* group not active */
              lookfor++;          /* next digit */
              break;
            }
          }
        }  /* for j<GROUPS */
        if ( !found ) {           /* no match! */
          total[NVOTE]++;
          repeat = FALSE;
        }
      }  /* for repeat */
    } /* for i<count */

    printf( "\nTotals: \n" );
    for( i=0; i<=NVOTE; i++ ) {
      if( active[i] ) {
        printf( "  %4d %s\n", total[i], gtxt[i] );
      }
    }
    printf( "\n" );

    /* sort votes... */
    for( i=0; i<=NVOTE; i++ ) idx[i] = i;
    for( i=0; i<=NVOTE; i++ )
      for( j=0; j<=NVOTE-1; j++ )
        if( total[idx[j]] > total[idx[j+1]] ) {
          k  = idx[j]; idx[j] = idx[j+1]; idx[j+1] = k;
        }

    k = ( ( count+1 ) )/2;
    if( total[idx[NVOTE]] >= k ) {  /* majority */
      printf( "    %s wins with a majority.\n", gtxt[idx[NVOTE]] );
      break;
    } else {
      printf( "    There is no majority.  Eliminate the least popular.\n" );
      for( i=0; i<NVOTE; i++ ) idx[i] = i;
      for( i=0; i<NVOTE; i++ )
        for( j=0; j<NVOTE-1; j++ )
          if( total[idx[j]] > total[idx[j+1]] ) {
            k  = idx[j]; idx[j] = idx[j+1]; idx[j+1] = k;
          }
      for( k=0; k<NVOTE; k++ )
        if( active[idx[k]] ) break;
      for( i=0; i<NVOTE; i++ ) {
        if( active[idx[i]] && total[idx[i]] == total[idx[k]] ) {
          printf( "      %s is eliminated from the choices.\n", gtxt[idx[i]] );
          active[idx[i]] = FALSE;
        }
      }
    }
  }                             /* for( ;; ) */
  return( idx[NVOTE] );
}

/* --------------------------------------------------------------------
/* Results for STV voting.
/* --------------------------------------------------------------------
*/
void result_stv()
{
  int loser[MAXGROUPS], order[MAXGROUPS], i, g;

  printf("%s results - %d valid votes\n\n", u_votename, count );

  printf( "For %d votes, %d votes are necessary for a majority.\n",
    count, ( count+1 )/2 );

  if( u_votetype == STVCFV_T ) {
    fprintf( stderr, "WARNING - STVCFV voting is treated exactly the same as\n" );
    fprintf( stderr, "          STV till a method is agreed on.\n" );
  }

  for( i=0; i<GROUPS; i++ ) {     /* all losers to start with */
    loser[i] = TRUE;
  }

  for( g=0; g<=NVOTE; g++ ) {
    for( i=0; i<GROUPS; i++ ) {     /* all losers to start with */
      active[i] = loser[i];
    }
    order[g] = tally_stv();
    loser[order[g]] = FALSE;      /* no longer a loser */
    if( !dorepeat ) break;
  }

  if( dorepeat ) {
    printf( "\nChoices in order of preference:\n" );
    for( g=0; dorepeat && g<=NVOTE; g++ ) {
      printf( "  %s\n",  gtxt[order[g]] );
    }
  }

}

/* --------------------------------------------------------------------
/* Results for non-stv voting.  Just format results, actually.
/* --------------------------------------------------------------------
*/
void result_nonstv()
{
  int i;

  /* Do table format output
  */
  if( dotable ) {
    printf("%s results - %d valid votes\n", u_votename, count );
    printf("\n Yes   No : 2/3? >100? : Pass? : Group\n");
    printf("---- ---- : ---- ----- : ----- : -------------------------------------------\n");
    for(i=0; i<GROUPS; i++) {
      cond1 = ( yesvote[i] >= (novote[i] * 2)) ? 1 : 0;
      cond2 = ( yesvote[i] >= (novote[i] + 100)) ? 1 : 0;
      printf("%4d %4d :  %s   %s :   %s : %s\n",
        yesvote[i], novote[i], noyes[cond1], noyes[cond2],
        noyes[ cond1 && cond2 ], groupstxt[i] );
    }
    if( GROUPS == 1 && absvote[0]>0 ) {
      if( badcount ) {
        printf( "%4d abstaining vote%c and %d invalid vote%c\n", absvote[0], absvote[0]!=1 ? 's' : ' ', badcount, badcount!=1 ? 's' : ' ' );
      } else {
        printf( "%4d abstaining vote%c\n", absvote[0], absvote[0]!=1 ? 's' : ' ' );
      }
    } else {
      if( badcount ) {
        printf( "%4d invalid vote%c\n", badcount, badcount!=1 ? 's' : ' ' );
      }
    }

  /* Do single vote "easy" format
  */
  } else {
    cond1 = ( yesvote[0] >= (novote[0] * 2)) ? 1 : 0;
    cond2 = ( yesvote[0] >= (novote[0] + 100)) ? 1 : 0;

    /* print and center summary line */
    sprintf( line, "%smoderated group %s %s %d:%d", u_moderated[0] ? "" : "un",
         groupstxt[0], failpass[ cond1 && cond2 ? 1 : 0 ], yesvote[0], novote[0] );
    i = (RIGHT-strlen(line))/2;
    if( i<0 ) i=0;                /* just in case too big... */
    while( i-- ) putc( ' ', stdout );
    printf( "%s\n\n", line );

    sprintf( line, "There were %d YES vote%c", yesvote[0], yesvote[0]!=1 ? 's':'#' );
    wrapout( line );
    sprintf( line, " and %d NO vote%c", novote[0], novote[0]!=1 ? 's':'#' );
    wrapout( line );
    sprintf( line, ", for a total of %d valid vote%c.", count, count!=1 ?'s':'#' );
    wrapout( line );
    if( absvote[0]>0 ) {          /* print abstain/invalid */
      if( absvote[0]>1 ) sprintf( line, "  There were %d abstains", absvote[0] );
                    else sprintf( line, "  There was 1 abstain" );
      wrapout( line );
      if( badcount ) sprintf( line, " and %d invalid ballot%c.", badcount, badcount!=1 ? 's':'#' );
                else sprintf( line, "." );
      wrapout( line );
    } else {
      if( badcount ) {
        if( badcount != 1 ) sprintf( line, "  There were %d invalid ballots.", badcount );
                     else sprintf( line, "  There was %d invalid ballot." );
        wrapout( line );
      }
    }
    wrapout( "\n" );

    wrapout( "For group passage, YES votes must be at least 2/3 of all valid ");
    wrapout( "(YES and NO) votes. " );
    wrapout( "  There also must be at least 100 more YES votes than NO votes. ");

    if( cond1 && cond2 ) {
      wrapout( "\nThere is a five day discussion period " );
      wrapout( "after these results are posted.  If no serious allegations of ");
      wrapout( "voting irregularities are raised, the moderator of news." );
      wrapout( "announce.newgroups will create the group shortly thereafter.\n");
    } else {
      wrapout( "\nThere is a five day discussion period " );
      wrapout( "after these results are posted.  Unless serious allegations of ");
      wrapout( "voting irregularities are raised, the group may not be voted " );
      wrapout( "on again for six months.\n");
    }
  }

  if( dodate ) {
    printf("\n");
    for(i=0; i<GROUPS; i++) {
      if( passdate[i] ) {
        printf("  %s passed on %s", groupstxt[i], passdate[i]);
      }
    }
  }
}

/* --------------------------------------------------------------------
/* Main procedure - This does everything
/* --------------------------------------------------------------------
*/

int main( argc, argv )
int argc;
char **argv;
{
  int i, j, arg;

  init_global_config();
  if( u_stv ) {
    votes = (char**)malloc(u_maxvotes*sizeof(char*));
    if( !votes ) {
      fprintf( stderr, "ERROR - not enough memory for vote array" );
      exit( 55 );
    }
  } else {
    novote = (int*)malloc(GROUPS*sizeof(int));
    yesvote = (int*)malloc(GROUPS*sizeof(int));
    absvote = (int*)malloc(GROUPS*sizeof(int));
    passdate = (char**)malloc(GROUPS*sizeof(char*));
    if( !passdate ) {
      fprintf( stderr, "ERROR - not enough memory for passdate array" );
      exit( 55 );
    }
  }

  fprintf( stderr, VERSION );
  if( argc<2 ) help();
  arg=1; dodate = dorepeat = FALSE;
  dotable = ( GROUPS != 1 );
  while( arg<argc && argv[arg][0] == '-' ) {
    if( argv[arg][1] == 'd' ) {
      dodate = TRUE;
      arg++;
    } else if( argv[arg][1] == 't' ) {
      dotable = TRUE;
      arg++;
    } else if( argv[arg][1] == 'r' ) {
      dorepeat = TRUE;
      arg++;
    } else {
      help();
    }
  }
  if( arg != argc-1) help();

  if ( !( fin = fopen( argv[arg], UV_OPEN_READ ))) {
    printf( "Can't open %s for input\n", argv[1] );
    exit( 2 );
  }


  if( u_stv ) {
    for(i=0; i<GROUPS; i++) {        /* all groups are active */
      active[i] = TRUE;
      strcpy( gtxt[i], groupstxt[i] );
    }
    strcpy( gtxt[NVOTE], "No Preference" );
    active[NVOTE] = TRUE;
  } else {
    for(i=0; i<GROUPS; i++) {        /* no votes yet! */
      novote[i] = yesvote[i] = absvote[i] = 0;
      passdate[i]=(char*)0;
    }
  }
  count = badcount = 0;

  for( ;; ) {
    if( !fgets( line, 200, fin ) ) {
      break;                      /* exit from loop */
    }
    for( j=0; j<5; j++ ) {
      if( j > 0 ) {
        if( !fgets( line, 200, fin ) ) {
          printf("Error!  Unexpected end of file!\n");
          exit(10+j);
        }
      }
      if( line[0]!=rtag[j]) {
        printf("Error!  %s not found: %s\n", rname[j], line);
        exit(10+j);
      }
      if( j == 2 ) strcpy(date, &line[3]);
    }

    if( line[3] == '!' ) {        /* bad vote! */
      badcount++;
      continue;
    }

    if( line[3] == '*' ) {        /* cancelled vote */
      continue;
    }

    if( u_stv ) {
      votes[count] = (char*)malloc(GROUPS*sizeof(char));
      if( !votes[count] ) {
        fprintf( stderr, "ERROR - not enough memory for vote %d\n", count );
        exit( 55 );
      }
      for(i=0; i<GROUPS; i++) votes[count][i] = line[i+3];
      count++;
    } else {
      if( GROUPS != 1 ) count++;
      for(i=0; i<GROUPS; i++) {
        switch( line[i+3] ) {
          case 'Y':
            yesvote[i]++;
            if( GROUPS == 1 ) count++;
            break;
          case 'N':
            novote[i]++;
            if( GROUPS == 1 ) count++;
            break;
          case '-':
            absvote[i]++;
            break;
          default:
            printf("Error!  Bad vote #%d %s\n", count, line);
            exit(20);
        }
        cond1 = ( yesvote[i] >= (novote[i] * 2)) ? 1 : 0;
        cond2 = ( yesvote[i] >= (novote[i] + 100)) ? 1 : 0;
        if( cond1 && cond2 ) {
          if( passdate[i]==0 ) passdate[i]=strdup(date);
        } else {
          if ( passdate[i] ) {
            free(passdate[i]);
          }
          passdate[i]=0;
        }
      }
    }
  }  /* for */
  fclose(fin);

  if( u_stv ) {
    result_stv();
  } else {
    result_nonstv();
  }

  return( 0 );
}

