/*---------------------------------------------------------------
/* Find Duplicates (c)1993,94 Ron Dippold All Rights Reserved
/*-------------------------------------------------------------*/
#define VERSION "\nUseVote 3.0 Duplicate Finder - (c) 1993,94 Ron Dippold\n\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usevote.h"

FILE *fin=NULL, *fout=NULL, *ftrace=NULL;
FILE *ftemp=NULL;                 /* non-memory sort */

char line[200];                   /* input line */
char addr[80], addr2[80];         /* name and address */
int count, linenum;
int deleted;
int dotrace;                      /* boolean flags */
int doask;                        /* ask for deletion */
#define NOASK 0                   /* never ask */
#define SOMEASK 1                 /* ask if there's a question */
#define YESASK 2

typedef struct {
  postype pos;                    /* position of vote in file */
  byte    del;                    /* deleted? */
  int     oidx;                   /* original index */
  char    *addr;                  /* memsort: pointer to A: info */
  char    *name;                  /* memsort: pointer to N: info */
  postype pos2;                   /* nonmemsort: position of addr in temp file */
} result_type;

result_type *item;
int *pback;

int  steps[6]={51,19,7,3,1,0};    /* for sorting */

char rtag[6] = "ANDCV";
char rname[5][8] = { "Address", "Name", "Date", "Comment", "Vote" };


#define MSG(x) fprintf( stderr, x )
int compare( );

/* ------------------------------------------------------------------
/* exit cleanly
/* ------------------------------------------------------------------
*/
void do_exit( code )
int code;                         /* code to exit with */
{
  int i;

  if( fin != NULL ) fclose( fin );
  if( fout != NULL ) fclose( fout );
  if( ftrace != NULL ) fclose( ftrace );
  if( u_memsort ) {
    for( i=0; i<u_maxvotes; i++ ) {
      if( item[i].addr ) free( item[i].addr ); /* free memory */
      if( item[i].name ) free( item[i].name ); /* free memory */
    }
  } else {
    if( ftemp != NULL ) fclose( ftemp );
  }
  exit( code );
}

/* -----------------------------------------------------------------
/* Compare string c1 with c2, ignore case.
/* c1<c2 = -1,  c1==c2 = 0,  c1>c2 = 1
/* -----------------------------------------------------------------
*/
int compare( c1, c2 )
char *c1;
char *c2;
{
  for( ;; ) {
    while( *c1 == '"') c1++;
    while( *c2 == '"') c2++;
    if( !*c1 ) {
      if( *c2 ) return( -1 );
                 else return( 0 );
    }
    if( !*c2 ) {
      if( *c1 ) return( 1 );
                 else return( 0 );
    }
    if( ( *c1 | 0x20 ) > ( *c2 | 0x20 ) )
      return( 1 );
    if( ( *c1 | 0x20 ) < ( *c2 | 0x20 ) )
      return( -1 );
    c1++; c2++;
  }
}

/* -----------------------------------------------------------
/* qsort callback routines; these do little more than call compare()
/* -----------------------------------------------------------
*/
int address_qsort_callback( r1 , r2 )
result_type *r1;
result_type *r2;
{
  static int i=0, j=3;
  static char spin[]="/-\\!";
  if( !( ++i % 100 ) ) {
    putc( spin[(++j)%4], stderr );  putc( '\b', stderr );
  }
  return( compare( r1->addr, r2->addr ));
}

int name_qsort_callback( r1 , r2 )
result_type *r1;
result_type *r2;
{
  static int i=0, j=3;
  static char spin[]="/-\\!";
  if( !( ++i % 100 ) ) {
    putc( spin[(++j)%4], stderr );  putc( '\b', stderr );
  }
  return( compare( r1->name, r2->name ));
}


/* ----------------------------------------------------
/* Do the sort
/* ----------------------------------------------------
*/
void sortaddr( isname )
int isname;        /* sort by name? */
{

  if( u_memsort ) {

    fprintf( stderr, "Memory Quicksorting %d votes...", count);
    qsort(item, count, sizeof(result_type),
          (isname ? name_qsort_callback : address_qsort_callback));
    MSG(" \n\n");

  } else {

    int i, pass, step, swapped, k, i1, j, which, start, end;
    result_type titem;
    char cmp[2][80], cvote[2];

    fprintf( stderr, "Disk Shellsorting %d votes...", count);
    pass = 0;

    do {
      if ( !(step = steps[pass]) ) {
        MSG("\n\n");
        return;                        /* all done */
      }
      pass++;
      fprintf( stderr, "\n  Pass %d: [", pass);
      for( i=0; i<step; i++ ) fputc( ' ', stderr );
      fputc( ']', stderr );
      for( i=0; i<=step; i++ ) fputc( '\b', stderr );

      for( start=0; start<step; start++) {
        end = count;
        MSG( ".");
        swapped = 1;
        while( start+step < end && swapped ) {
          i1 = start;
          which=0;
          fseek( ftemp, item[i1].pos2, SEEK_SET );
          fgets( cmp[which], 80, ftemp ); /* read address */
          if( isname ) {
            fgets( cmp[which], 80, ftemp ); /* read name */
          }
          swapped = 0;

          for( j=start+step; j<end; j+=step) {
            fseek( ftemp, item[j].pos2, SEEK_SET );
            fgets( cmp[which^1], 80, ftemp ); /* read address */
            if( isname ) {
              fgets( cmp[which^1], 80, ftemp ); /* read name */
            }

            k = compare( cmp[which], cmp[which^1]);

            if ( k > 0) {   /* swap i and j */
              titem = item[i1]; item[i1] = item[j]; item[j] = titem;
              swapped = 1;
            } else {
              which = which^1;                 /* other buffer */
            }
            i1 = j;

          }  /* for j= */
          end -= step;
        }  /* do while */
      } /* for start */
    }  while (1); /* do */
    MSG("\n\n");
  }
}

/* ----------------------------------------------------
/* Format vote for trace file
/* ----------------------------------------------------
*/
void tformat( line, desc, addr, name, vote )
char *line, *desc, *addr, *name, *vote;
{
  int i, j;

  i=0;
  for(j=0; desc[j]; line[i++] = desc[j++] );
  line[i++]=' ';
  if( vote[3] == '!' || vote[3] == '*' ) {
    for(j=0; j<GROUPS; j++) line[i++] = vote[3];
  } else {
    for(j=0; j<GROUPS && vote[3+j]!='\n'; line[i++] = vote[3+j++] );
  }
  line[i++]=' ';
  for(j=3; addr[j]!='\n'; line[i++] = addr[j++] );
  if( name[3] !='\n' && name[3] ) {
    line[i++]=' ';
    line[i++]='(';
    for(j=3; name[j]!='\n' && j<75 ; line[i++] = name[j++] );
    line[i++]=')';
  }
  line[i++]='\0';
}


/* ----------------------------------------------------
/* Ask about deleting duplicate
/*  isname is true if doing name check
/* ----------------------------------------------------
*/
void askdel( isname, j, i )
int isname;
int j;
int i;
{
  int old, new, k;
  char c;
  static char vote1[5][80], vote2[5][80];

  if( item[j].pos < item[i].pos ) {
    old = j; new = i;
  } else {
    old = i; new = j;
  }
  fseek( fin, item[old].pos, SEEK_SET );
  for( k=0; k<5; k++ ) fgets( vote1[k], 80, fin );
  fseek( fin, item[new].pos, SEEK_SET );
  for( k=0; k<5; k++ ) fgets( vote2[k], 80, fin );
  if( doask == YESASK ) {         /* always ask */
    c = ' ';
  } else {
    c=' ';                        /* it isn't pretty, but easy to fix */
    if( !isname ) {               /* comparing addresses */
      if( compare( vote1[0], vote2[0] ) == 0  || /* exact addresses */
          compare( vote1[1], vote2[1] ) == 0 ) {  /* exact name */
        c='+';
      }
    } else {                      /* comparing names */
      if( compare( vote1[2], vote2[2] ) == 0 ) { /* exact addresses */
        c='+';
      }
    }
    if( doask == NOASK ) c = '+';
    if( c == '+' ) {              /* do automatically */
      c='O';
      if( vote2[4][3] == '!' ) c = '*';
      if( vote1[4][3] == '!' ) c = 'O';
    }
  }
  if( c == ' ' ) {
    printf( "\n\n* %d%% Progress\n", (( i>j ) ? i : j )*100/count );
    printf("Older vote:\n" );
    for( k=0; k<5; k++ ) printf( " %s", vote1[k] );
    printf("\nNewer vote:\n");
    for( k=0; k<5; k++ ) printf( " %s", vote2[k] );
    printf("\nThese votes might be duplicates.  Or not.\n");
    while( c != 'O' && c != 'N' && c!='*') {
      printf( "Delete O)lder vote, N)either, *)Newer vote? " );
      fgets( line, 200, stdin );
      c = toupper( line[0] );
    }
  }
  if( c == 'O' ) {
    deleted++;
    item[old].del = TRUE;
    if( dotrace ) {
      tformat( line, "Deleted:", vote1[0], vote1[1], vote1[4] );
      fprintf( ftrace, "%s\n", line );
      tformat( line, "   Kept:", vote2[0], vote2[1], vote2[4] );
      fprintf( ftrace, "%s\n\n", line );
    }
  }
  if( c == '*' ) {
    deleted++;
    item[new].del = TRUE;
    if( dotrace ) {
      tformat( line, "Deleted:", vote2[0], vote2[1], vote2[4] );
      fprintf( ftrace, "%s\n", line );
      tformat( line, "   Kept:", vote1[0], vote1[1], vote1[4] );
      fprintf( ftrace, "%s\n\n", line );
    }
  }
}

void help()
{
  MSG( " uvdup (-y) (-n) <old results> <new results> (<trace file>)\n\n" );
  MSG( " This program will look through voting result file <old results> for\n");
  MSG( " possible duplicate votes.  You will be given the option to erase\n");
  MSG( " the older duplicate vote.  The final file should have no duplicates.\n\n");
  MSG( " If you give a trace file name, a list of all votes deleted as duplicates\n" );
  MSG( " will be printed to that file.  Useful to see what got axed.\n\n" );
  MSG( " -n will cause uvdup to be paranoid - it won't delete anything without\n" );
  MSG( "   asking you first.  Normally it only asks if there's a question.\n" );
  MSG( " -y will cause uvdup to automatically remove anything it even suspects\n" );
  MSG( "   is a duplicate vote without asking you about it. You should NOT\n" );
  MSG( "   normally do this.  -y and -n are mutually exclusive. \n\n" );

  exit( 1 );
}


/* Check for line too long */
void checklen()
{
  int i;
  for(i=0; i<80 && line[i]; i++);
  if( i == 80 ) {
    line[79] = 0;
    fprintf(stderr,"Error - line %d is too long.\n%s\n", linenum, line);
    exit(44);
  }
}


/* ----------------------------------------------------
/* Main procedure - almost everything goes here
/* ----------------------------------------------------
*/
int main( argc, argv )
int argc;
char **argv;
{
  int i,j,i1,j1,k, args;
  long mtot=0;

  init_global_config();
  item = (result_type*)malloc(u_maxvotes*sizeof(result_type));
  if( !item ) {
    fprintf( stderr, "Non enough memory for item array!\n" );
    exit( 60 );
  }
  pback = (int*)malloc(u_maxvotes*sizeof(int));
  if( !pback ) {
    fprintf( stderr, "Non enough memory for pback array!\n" );
    exit( 61 );
  }

  MSG( VERSION );

  dotrace = FALSE;                /* don't trace */
  doask = SOMEASK;                /* ask sometime */

  args=0;
  if( argc > 1 ) {
    if( argv[args+1][0] == '-' ) {
      if( argv[args+1][1] == 'y' ) {
        args++;
        doask=NOASK;
      } else if ( argv[args+1][1] == 'n' ) {
        args++;
        doask=YESASK;
      } else {
        help();
      }
    }
  }

  if( argc-args !=3 && argc-args !=4 ) {
    help();
  }

  if( argv[args+1][0] == '-' ) help(); /* no two options */

  if( argc-args == 4 ) {
    dotrace = TRUE;
  }

  if( u_memsort ) {
    for( j=0; j<u_maxvotes; item[j++].addr = NULL ); /* null out pointers */
    for( j=0; j<u_maxvotes; item[j++].name = NULL ); /* null out pointers */
  }

  if ( !( fin = fopen( argv[args+1], UV_OPEN_READ ))) {
    fprintf( stderr, "Can't open %s for input\n", argv[args+1] );
    exit( 2 );
  }

  if ( ( fout = fopen( argv[args+2], UV_OPEN_READ ))) {
    int c;
    fclose( fout );
    printf( "\n WARNING! File %s already exists.\n", argv[args+2] );
    printf( "Do you really want to destroy it? " );
    fgets( line, 80, stdin );
    c = line[0];
    if( c !='y' && c !='Y' ) {
      do_exit( 20 );
    }
  }

  if( !u_memsort ) {
    if ( !( ftemp = fopen( argv[args+2], UV_OPEN_WRITE ))) {
      fprintf( stderr, "Can't open %s for output\n", argv[args+2] );
      do_exit( 2 );
    }
  }

  if( dotrace ) {
    if ( !( ftrace = fopen( argv[args+3], UV_OPEN_WRITE ))) {
      fprintf( stderr, "Can't open %s for output\n", argv[args+3] );
      do_exit( 3 );
    }
  }


  /* ------------------- Read all the votes for sorting ------------------- */

  count = linenum = deleted = 0;
  for( ;; ) {
    if( count >= u_maxvotes ) {
      fprintf(stderr, "  ERROR!  Exceeded maximum of %d votes.  Check usevote.cfg\n",u_maxvotes);
      do_exit(30);
    }
    item[count].oidx = count;
    item[count].pos = ftell( fin );
    if( !u_memsort ) {
      item[count].pos2 = ftell( ftemp );
    }
    linenum++;
    if( !fgets( line, 200, fin ) ) {
      break;                      /* exit from loop */
    }
    checklen();
    if( line[0]!='A') {
      fprintf( stderr, "Error!  Address not found line %d\n", linenum);
      do_exit(9);
    }
    /* strip anything over two periods worth off address */
    k=0;
    for(i=3; line[i] && line[i] != '@'; i++ );
    while( line[i] ) {
      if( line[i++] == '.' ) k++;
    }
    if( k > 1 ) {
      for(i=3,j=3; line[i] != '@'; i++ ) {
        line[j++] = line[i];
      }
      line[j++] = '@';
      do {
        while( line[i++] != '.' );
      } while( --k != 1 );
      while( line[i] ) {
        line[j++] = line[i++];
      }
      line[j]='\0';
    } else {
      j=i;
    }
    if( u_memsort ) {
      line[j-1]='\0';               /* no lf, thanks */
      j-=3;
      mtot += j;
      if( (item[count].addr = ( char * ) malloc( j )) == NULL ) {
        fprintf( stderr, "  Not enough memory for qsort...  %ld\n%s",mtot,
          "  Try recompiling with MEMSORT disabled in usevote.h.\n" );
        do_exit(99);
      } else {
        memcpy( item[count].addr, &line[3], j );
      }
    } else {
      fputs( &line[3], ftemp );           /* print to temp file */
    }

    linenum++;
    if( !fgets( line, 200, fin ) ) {
      fprintf( stderr, "Error!  Unexpected end of file line %d!\n", linenum);
      do_exit(10);
    }
    checklen();
    if( line[0]!='N') {
      fprintf( stderr, "Error!  Name not found line %d\n", linenum);
      do_exit(10);
    }
    if( u_memsort ) {
      for( j=0; line[j]!='\n'; j++ ); /* no lf... */
      while( j<3 ) line[j++]=' ';
      line[j++]=0;
      j-=3;
      mtot += j;
      if( (item[count].name = ( char * ) malloc( j )) == NULL ) {
        fprintf( stderr, "  Not enough memory for qsort.  %ld\n%s",mtot,
          "  Try recompiling with MEMSORT disabled in usevote.h.\n");
        do_exit(99);
      } else {
        memcpy( item[count].name, &line[3], j );
      }
    } else {
      fputs( &line[3], ftemp );           /* print to temp file */
    }
    for( i=0; i<3; i++ ) {
      linenum++;
      if( !fgets( line, 200, fin ) ) {
        fprintf( stderr, "Error!  Unexpected end of file line %d!\n", linenum);
        do_exit(11);
      }
      checklen();
      if( line[0]!=rtag[i+2]) {
        fprintf( stderr, "Error!  %s not found line %d\n", rname[i+2], linenum);
        do_exit(11);
      }
    }
    item[count].del=FALSE;
    count++;
  }


  if( !u_memsort ) {
    fclose( ftemp );

    if ( !( ftemp = fopen( argv[args+2], UV_OPEN_READ ))) {
      fprintf( stderr, "Can't open %s as temporary file\n", argv[args+2] );
      fclose( fin );
      do_exit( 2 );
    }
  }

/* ------------------------ Do sorting by address ----------------- */

  MSG("By Address: ");
  if( count>1 ) sortaddr( 0 );

  j = 0;
  if( u_memsort ) {               /* memory compare */
    for( i=1; i<count; i++ ) {
      if( compare( item[j].addr, item[i].addr ) == 0 ) {
        while( ++i<count ) {        /* find end of this set of duplicates */
          if( compare( item[j].addr, item[i].addr ) != 0 ) break;
        }
        for( i1=j+1; i1<i; i1++ )
          for( j1=j; j1<i1; j1++ )
            if( !item[j1].del && !item[i1].del ) askdel( FALSE, j1, i1 );
      }
      j = i;
    }
  } else {                        /* disk compare */
    fseek( ftemp, item[j].pos2, SEEK_SET );
    fgets( addr, 80, ftemp );
    for( i=1; i<count; i++ ) {
      fseek( ftemp, item[i].pos2, SEEK_SET );
      fgets( addr2, 80, ftemp );
      if( compare( addr, addr2 ) == 0 ) {
        while( ++i<count ) {        /* find end of this set of duplicates */
          fseek( ftemp, item[i].pos2, SEEK_SET );
          fgets( addr2, 80, ftemp );
          if( compare( addr, addr2 ) != 0 ) break;
        }
        for( i1=j+1; i1<i; i1++ )
          for( j1=j; j1<i1; j1++ )
            if( !item[j1].del && !item[i1].del ) askdel( FALSE, j1, i1 );
      }
      j = i;
      strcpy( addr, addr2 );
    }
  }


/* ------------------------ Do sorting by name ----------------- */

  MSG("\n\nBy Name: ");
  if( count>1 ) sortaddr( 1 );

  for( j=0; item[j].del && j<count; j++ ); /* first non-deleted */
  for( i=j+1; item[i].del && i<count; i++ ); /* second non-deleted */

  if( u_memsort ) {               /* memory compare */
    for( ;i<count; i++ ) {
      if( item[i].del ) continue;
      if( item[i].name[0] ) {
        if( compare( item[j].name, item[i].name ) == 0 ) {
          while( ++i<count ) {        /* find end of this set of duplicates */
            if( !item[i].del ) {
              if( compare( item[j].name, item[i].name ) != 0 ) break;
            }
          }
          for( i1=j+1; i1<i; i1++ )
            for( j1=j; j1<i1; j1++ )
              if( !item[j1].del && !item[i1].del ) askdel( TRUE, j1, i1 );
        }
      }
      j = i;
    }
  } else {                        /* disk compare */
    fseek( fin, item[j].pos, SEEK_SET );
    fgets( addr, 80, fin );
    fgets( addr, 80, fin );
    for( ; i<count; i++ ) {
      if( item[i].del ) continue;
      fseek( fin, item[i].pos, SEEK_SET );
      fgets( addr2, 80, fin );
      fgets( addr2, 80, fin );
      if( addr2[3] != '\n' ) {
        if( compare( addr, addr2 ) == 0 ) {
          while( ++i<count ) {        /* find end of this set of duplicates */
            if( !item[i].del ) {
              fseek( fin, item[i].pos, SEEK_SET );
              fgets( addr2, 80, fin );
              fgets( addr2, 80, fin );
              if( compare( addr, addr2 ) != 0 ) break;
            }
          }
          for( i1=j+1; i1<i; i1++ )
            for( j1=j; j1<i1; j1++ )
              if( !item[j1].del && !item[i1].del ) askdel( TRUE, j1, i1 );
        }
      }
      j = i;
      strcpy( addr, addr2 );
    }
    fclose( ftemp );
  }

  MSG("\n\nCopying results to new result file.\n");
  if ( !( fout = fopen( argv[args+2], UV_OPEN_WRITE ))) {
    fprintf( stderr, "Can't open %s as output\n", argv[args+2] );
    do_exit( 2 );
  }

  for( i=0; i<count; i++ ) {      /* sort by original index */
    pback[item[i].oidx] = i;
  }

  for( i=0; i<count; i++ ) {
    if( item[pback[i]].del ) continue;
    fseek( fin, item[pback[i]].pos, SEEK_SET );
    for( j=0; j<5; j++ ) {
      fgets( addr, 80, fin );   fputs( addr, fout );
    }
  }

  if ( dotrace ) fclose( ftrace );
  fclose( fout );
  fclose( fin );

  fprintf( stderr, "\nAll done, %d vote(s) deleted.\n", deleted );
  return( 0 );
}
