/*---------------------------------------------------------------
/* Process Voting  (c)1993,94 Ron Dippold All Rights Reserved
/* UseVote Version 2.2
/*
/* This software processes the votes from a mail file or a copy
/* of votes extracted from your mail.  It needs to follow the
/* standard header format with "From:" etc.
/*-------------------------------------------------------------*/
#define VERSION "UseVote 2.2 Vote Processing - (c) 1993,94 Ron Dippold"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usevote.h"

FILE *fin=NULL, *fcmd=NULL, *fvotes=NULL, *ftrace=NULL;

#define LINESIZE 4000
char line[LINESIZE];              /* input line */
char mline[LINESIZE];             /* get line from file */
static char uin[81];              /* user input */
char name[80], addr[80];          /* name and address */
char date[80], comment[80];       /* date and comment */
int count;
int dotrace;
int dotest;                       /* just test stuff */
int dointer;                      /* interactive? */
int startisend;                   /* if start is same as end for mail */
int lmailstart, lmailend;

postype mpos;                        /* message starting position */
int mlines;                       /* message lines */

char vmark;                       /* found votemark */

int groupslen[MAXGROUPS];          /* length of string */
int votes[MAXGROUPS];              /* the votes for this ballot */
int mvotes[MAXGROUPS];

struct uv_s_rules {
  char *test1;
  char *test2;
} *rules;
int nrules;                       /* number of rules */

int dateslen[DTXTS];              /* length of date header string */
char datefnd[DTXTS][80];          /* dates found in message */

#define RULECHAR "YNAV.ynav"

/* define state machine states */
#define NOTHING_S    0            /* no state */
#define NEEDNAME_S   1            /* need voter name */
#define NEEDBALLOT_S 2            /* looking for ballot */
#define BALLOT_S     3            /* in ballot */
int state= NOTHING_S;

/* define return code for interact() */
#define CONTINUE_C  0             /* continue with error */
#define RETRY_C     1             /* try processing again */
#define RETRYB_C    2             /* retry with ballot found forced */
#define DISCARD_C   3             /* discard the vote */
#define ADDROK_C    4             /* address is ok */


#ifndef PAGEWIDE
#define PAGEWIDE 80
#endif
#ifndef PAGELEN
#define PAGELEN 24
#endif
#define PAGELINES ( PAGELEN-5 )

void do_exit();
int rule_check1();
int rule_check();
void rule_printg();


/* ------------------------------------------------------------------
/* exit cleanly
/* ------------------------------------------------------------------
*/
void do_exit( code )
int code;                         /* code to exit with */
{
  if( fin != NULL ) fclose( fin );
  if( fcmd != NULL ) fclose( fcmd );
  if( fvotes != NULL ) fclose( fvotes );
  if (dotrace) if( ftrace != NULL) fclose( ftrace );
  exit( code );
}


/* ------------------------------------------------------------------
/* Print a list of which groups match the criteria
/* ------------------------------------------------------------------
*/
void rule_printg( f, test, c, text, do_or )
FILE  *f;                         /* where to print it */
char  *test;                      /* string holding test */
char  c;                          /* which character to look for */
char  *text;                      /* description of character */
int   do_or;                      /* if true, OR, otherwise AND */
{
  static int firstrule;           /* if 1, first rule line */
  int i, cnt;

  if( test == NULL ) {            /* reset the thing */
    firstrule = TRUE;
    return;
  }
  for( cnt=0, i=0; i<GROUPS; i++ ) { /* count number of matches */
    if( test[i] == c ) cnt++;
  }
  if( cnt == 0 ) {                 /* nothing found */
    return;
  }
  strcpy( mline, "    " );
  if( firstrule ) {
    firstrule = FALSE;
  } else {
    strcat( mline, "and " );
  }
  strcat( mline, text );
  fprintf( f, "%s ", mline );
  for( i=0; mline[i]; mline[i++]=' ' );
  if( do_or ) {
    mline[i-2] = '\0';
    strcat( mline, "or " );
  } else {
    mline[i-3] = '\0';
    strcat( mline, "and " );
  }
  for( i=0; i<GROUPS; i++ ) {
    if( test[i] == c ) {
      fprintf( f, "%s\n", groupstxt[i] );
      if( --cnt != 0 ) {
        fprintf( f, "%s", mline );
      }
    }
  }
}




/* ------------------------------------------------------------------
/* Print a verbal description of a rule to a given file
/* ------------------------------------------------------------------
*/
void rule_print( rule, f )
int   rule;                     /* which rule */
FILE  *f;                       /* where to print it */
{
  fprintf( f, "Voting Rule #%d:\n", rule+1 );
  fprintf( f, "  If you\n" );
  rule_printg( f, NULL, ' ', "", FALSE );    /* reset */
  rule_printg( f, rules[rule].test1, 'Y', "vote Yes on", FALSE );
  rule_printg( f, rules[rule].test1, 'N', "vote No on", FALSE );
  rule_printg( f, rules[rule].test1, 'A', "abstain on", FALSE );
  rule_printg( f, rules[rule].test1, 'V', "vote Yes or No on", FALSE );
  rule_printg( f, rules[rule].test1, 'y', "vote Yes on", TRUE );
  rule_printg( f, rules[rule].test1, 'n', "vote No on", TRUE );
  rule_printg( f, rules[rule].test1, 'a', "abstain on", TRUE );
  rule_printg( f, rules[rule].test1, 'v', "vote Yes or No on", TRUE );
  fprintf( f, "  Then you must\n" );
  rule_printg( f, NULL, ' ', "", FALSE );    /* reset */
  rule_printg( f, rules[rule].test2, 'Y', "vote Yes on", FALSE );
  rule_printg( f, rules[rule].test2, 'N', "vote No on", FALSE );
  rule_printg( f, rules[rule].test2, 'A', "abstain on", FALSE );
  rule_printg( f, rules[rule].test2, 'V', "vote Yes or No on", FALSE );
  rule_printg( f, rules[rule].test2, 'y', "vote Yes on", TRUE );
  rule_printg( f, rules[rule].test2, 'n', "vote No on", TRUE );
  rule_printg( f, rules[rule].test2, 'a', "abstain on", TRUE );
  rule_printg( f, rules[rule].test2, 'v', "vote Yes or No on", TRUE );
  fprintf( f, "\n\n" );
}


/* ------------------------------------------------------------------
/* This checks an individual rules test clause for compliance.  it is
/* used as a subroutine of rule_check().  Returns TRUE if the vote
/* matches the conditions in the clause.
/* ------------------------------------------------------------------
*/
int rule_check1( clause )
char *clause;                     /* test clause */
{
  int i;
  static int check[MAXGROUPS];
  static int passy, passn, passa, passv;

  passy = passn = passa = passv = FALSE;

  for( i=0; i<GROUPS; check[i++]=FALSE ); /* reset to all false */

  for( i=0; i<GROUPS; i++ ) {
    switch( clause[i] ) {
      case 'Y':
        if( votes[i] == YES_V ) check[i] = TRUE;
        break;
      case 'N':
        if( votes[i] == NO_V ) check[i] = TRUE;
        break;
      case 'A':
        if( votes[i] == ABSTAIN_V ) check[i] = TRUE;
        break;
      case 'V':
        if( votes[i] != ABSTAIN_V ) check[i] = TRUE;
        break;
      case '.':
        check[i] = TRUE;
        break;
      case 'y':
        if( votes[i] == YES_V ) passy = TRUE;
        break;
      case 'n':
        if( votes[i] == NO_V ) passn = TRUE;
        break;
      case 'a':
        if( votes[i] == ABSTAIN_V ) passa = TRUE;
        break;
      case 'v':
        if( votes[i] != ABSTAIN_V ) passv = TRUE;
        break;
    }
  }

  for( i=0; i<GROUPS; i++ ) {     /* now second pass */
    if( passy ) if( clause[i] =='y' ) check[i] = TRUE;
    if( passn ) if( clause[i] =='n' ) check[i] = TRUE;
    if( passa ) if( clause[i] =='a' ) check[i] = TRUE;
    if( passv ) if( clause[i] =='v' ) check[i] = TRUE;
  }

  for( i=0; i<GROUPS; i++ ) {
    if( check[i] == FALSE ) {
      return( FALSE );
    }
  }

  return( TRUE );
}



/* ------------------------------------------------------------------
/* Check a vote for compliance with the rules
/* -1 indicates compliance, anything else is the rule that failed
/* ------------------------------------------------------------------
*/
int rule_check( )
{
  int i;

  for( i=0; i<nrules; i++ ) {     /* for each rule */
    if( rule_check1( rules[i].test1 ) == TRUE ) {
      if( rule_check1( rules[i].test2 ) == FALSE ) {
        return( i );              /* failed! */
      }
    }
  }
  return( -1 );
}


/* ------------------------------------------------------------------
/* Read the rules from the rules file, check them out
/* ------------------------------------------------------------------
*/
void rule_read( )
{
  FILE *fr;
  int i, j, k;
  char c;

  if ( !( fr = fopen( u_rulesfile, UV_OPEN_READ ))) {
    printf( "Can't open rules file %s.\n", u_rulesfile );
    do_exit( 30 );
  }
  nrules = 0;
  while( fgets( mline, LINESIZE, fr ) ) {
    if( mline[0] == '#' ) continue; /* # is a comment */
    if( mline[0] == '\n' ) continue;
    i=0;
    while( space(mline[i]) ) i++;
    if( strlook( &mline[i], "if ", 3 ) != 0 ) {
      fprintf( stderr, "  Bad rule - 'if'? : %s\n", mline );
      fclose( fr ); do_exit( 31 );
    }
    i+=3;
    while( space(mline[i]) ) i++;
    for( j=0, k=0; !space(mline[i]); i++, j++ ) {
      if( j == GROUPS ) {
        fprintf( stderr, "  'if' clause too long: %s\n", mline );
        fclose( fr ); do_exit( 31 );
      }
      rules[nrules].test1[j] = c = mline[i];
      if( c!='Y' && c!='N' && c!='A' && c!='V' && c!='.' &&
          c!='y' && c!='n' && c!='a' && c!='v' ) {
        fprintf( stderr, "  Bad 'if' char '%c': %s\n", mline[i], mline );
        fclose( fr ); do_exit( 31 );
      }
      if( c!='.' ) k=1;
    }
    if( j != GROUPS ) {
      fprintf( stderr, "  'if' clause too short: %s\n", mline );
      fclose( fr ); do_exit( 31 );
    }
    if( k== 0 ) {
      fprintf( stderr, "  'if' clause is null: %s\n", mline );
      fclose( fr ); do_exit( 31 );
    }
    while( space(mline[i]) ) i++;
    if( strlook( &mline[i], "then ", 5 ) != 0 ) {
      fprintf( stderr, "  Bad rule - 'then'? : %s\n", mline );
      fclose( fr ); do_exit( 31 );
    }
    i+=5;
    while( space(mline[i]) ) i++;
    for( j=0, k=0; !space(mline[i]); i++, j++ ) {
      if( j == GROUPS ) {
        fprintf( stderr, "  'then' clause too long: %s\n", mline );
        fclose( fr ); do_exit( 31 );
      }
      rules[nrules].test2[j] = c = mline[i];
      if( c!='Y' && c!='N' && c!='A' && c!='V' && c!='.' &&
          c!='y' && c!='n' && c!='a' && c!='v' ) {
        fprintf( stderr, "  Bad 'then' char '%c': %s\n", mline[i], mline );
        fclose( fr ); do_exit( 31 );
      }
      if( c!='.' ) k=1;
    }
    if( j != GROUPS ) {
      fprintf( stderr, "  'then' clause too short: %s\n", mline );
      fclose( fr ); do_exit( 31 );
    }
    if( k== 0 ) {
      fprintf( stderr, "  'then' clause is null: %s\n", mline );
      fclose( fr ); do_exit( 31 );
    }
    if( dotest ) {
      rule_print( nrules, stdout );          /* print the rule */
    }
    nrules++;
  }
  fclose( fr );
}



/* ----------------------------------------------------------------------
/* Print head and/or body of message to ack file
/* ----------------------------------------------------------------------
*/
void headbody( fout, prefix, dohead )
FILE *fout;                       /* file to write out to */
char *prefix;                     /* to write before each line */
int dohead;                       /* true to print header too */
{
  postype cpos;                      /* current position */
  int lines;                      /* number of lines */
  int i, v, donehead, len;

  cpos = ftell( fin );
  fseek( fin, mpos, SEEK_SET );   /* start of message */
  fgets( mline, LINESIZE, fin );        /* skip start line */
  lines = 1;

  /* do header */
  for( donehead=FALSE; !donehead; lines++) {
    fgets( mline, LINESIZE, fin );      /* skip start line */
    donehead=TRUE;
    for( i=0; mline[i]; i++ ) {    /* look for blank line */
      if( !space( mline[i] ) ) {
        donehead=FALSE;
        break;
      }
    }
    if( dohead ) {                /* print header? */
      fprintf( fout, "%s%s", prefix, mline );
    } else {
      len = strlen( mline );
      for( v=0; v<GROUPS; v++ ) {     /* check for each group name */
        for( i=0; i<len-groupslen[v]; i++ ) {
          if( !isletter( mline[i] ) ) continue;
          if( strlook( groupstxt[v], &mline[i], groupslen[v] ) == 0 ) {
            if (isletter(mline[i+groupslen[v]] ) ) continue;
            if( mline[i+groupslen[v]] == '.' && isletter( mline[i+groupslen[v]+1] ))
              continue;
            fprintf( fout, "%s%s", prefix, mline );
            v = GROUPS;
            break;
          }
        }
      }
    }


  }

  /* do body */
  while ( lines++ < mlines ) {
    fgets( mline, LINESIZE, fin );
    fprintf( fout, "%s%s", prefix, mline );
  }

  fseek( fin, cpos, SEEK_SET );   /* return to last postion */
}





/* ----------------------------------------------------------------------
/* This grabs a section from the the acktext file and prints it to the
/* message file.
/* ----------------------------------------------------------------------
*/
void ackmsg( fout, tag )
FILE *fout;                       /* file to write out to */
char *tag;                        /* tag to look for in acktext file */
{

  static FILE *ftxt;
  static word taglen;
  static int didbody, i;
  static char pfx[80];

  /* Read text from ballot ack file and copy to individual ack */
  if ( !( ftxt = fopen( u_acktxtfile, UV_OPEN_READ ))) {
    fprintf( stderr, "Can't open %s for input!\n", u_acktxtfile );
    do_exit( 3 );
  }

  taglen = strlen( tag );

  /* find the tag text */
  for( ;; ) {
    if( !fgets( mline, LINESIZE, ftxt ) ) {
      fprintf( stderr, "Can't find section %s in %s!\n", tag, u_acktxtfile );
      fclose( ftxt ); do_exit( 31 );
    }
    if( strncmp( mline, tag, taglen ) == 0 ) {
      break;
    }
  }

  /* now transfer the tag text to the message file */
  for( ;; ) {
    if( !fgets( mline, LINESIZE, ftxt ) ) {
      fprintf( stderr, "Can't find [end] in %s!\n", u_acktxtfile );
      fclose( ftxt ); do_exit( 32 );
    }
    if( mline[0] == '#' ) {
      fprintf( stderr, "*** Found '#' in %s - missing '[end]'?\n", u_acktxtfile );
      fclose( ftxt ); do_exit( 33 );
    }
    if( strncmp( mline, "[end]", 5 ) == 0 ) {
      break;
    }

    if ( !dotest ) {
      for( didbody=FALSE, i=0; mline[i]; i++ ) {
        if( mline[i] == '%' ) {
          if( strncmp( "%body%", &mline[i], 6 ) == 0 ) { /* include body */
            for( pfx[i--]=0; i>=0; i-- ) pfx[i] = mline[i];
            didbody = TRUE;
            headbody( fout, pfx, FALSE );
            break;
          }
          if( strncmp( "%headbody%", &mline[i], 10 ) == 0 ) { /* include both */
            for( pfx[i--]=0; i>=0; i-- ) pfx[i] = mline[i];
            didbody = TRUE;
            headbody( fout, pfx, TRUE );
          }
        }
      }
    }

    if( !didbody ) {
      if( strncmp( mline, MAILSTART, lmailstart ) == 0 ) {
        fprintf( fout, "~" );       /* "From " not allowed in message */
      }
      fputs( mline, fout );
    }
  }
  fclose( ftxt );
}


/* -----------------------------------------------------------------
/* printvote - print out how this particular vote looks
/* -----------------------------------------------------------------
*/
void printvote( fout )
FILE *fout;                       /* output file */
{
  int i;

  for( i=0; i<GROUPS; i++ ) {
    if( u_stv ) {
      if( votes[i] == ABSTAIN_V ) fprintf( fout, "     %s\n", groupstxt[i] );
                             else fprintf( fout, "   %c %s\n", votes[i], groupstxt[i] );
    } else {
      if( votes[i] == YES_V )     fprintf( fout, "     YES vote on %s\n", groupstxt[i] );
      if( votes[i] == NO_V )      fprintf( fout, "      NO vote on %s\n", groupstxt[i] );
      if( votes[i] == ABSTAIN_V ) fprintf( fout, " ABSTAIN vote on %s\n", groupstxt[i] );
    }
  }
  fprintf( fout, "\n" );
}

/* -----------------------------------------------------------------
/* convert votes to vote string
/* -----------------------------------------------------------------
*/
void votes_str( votes, mvotes, str )
int *votes;                       /* the votes */
int *mvotes;                      /* number of choices for vote */
char *str;                        /* string to put votes in */
{
  int i;

  for( i=0; i<GROUPS; i++ ) {
    if( mvotes[i]>1 ) {
      str[i]='!';
    } else {
      switch(votes[i]) {
        case NO_V:
          str[i]='N';
          break;
        case YES_V:
          str[i]='Y';
          break;
        case ABSTAIN_V:
          str[i]='-';
          break;
        case CANCEL_V:
          str[i]='C';
          break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
          str[i]=( char ) votes[i];
          break;
        default:
          str[i]='!';
          break;
      }
    }
  }
  str[i]=0;
}

/* -----------------------------------------------------------------
/* write vote to votes file
/* -----------------------------------------------------------------
*/
void writevote( votes )
char *votes;                      /* Votes to write for V:  */
{
  fprintf( fvotes, "A: %s\n", addr ); /* save address */
  fprintf( fvotes, "N: %s\n", name ); /* save name */
  fprintf( fvotes, "D: %s\n", date ); /* save date */
  fprintf( fvotes, "C: %s\n", comment ); /* comment */
  fprintf( fvotes, "V: %s\n", votes ); /* votes ( or error ) */
}


/* ---------------------------------------------------------------------
/* Show part of a message - returns the number of lines left
/* -----------------------------------------------------------------
*/
int showmsg( begin, len )
int begin;                        /* start line in message */
int len;                          /* number of lines to show */
{
  static char str[PAGEWIDE+1];
  postype cpos;                      /* current position */
  int lines;                      /* number of lines */
  int plines;                     /* printed lines */
  int i, j;

  cpos = ftell( fin );            /* save current position */
  fseek( fin, mpos, SEEK_SET );   /* start of message */

  for( lines=0; lines<begin; lines++ ) {
    fgets( mline, LINESIZE, fin );/* lines */
  }

  for( plines=0; plines<len; ) {
    if( lines >= mlines ) {      /* past end */
      strcpy( mline, "\n" );       /* blank line */
    } else {
      fgets( mline, LINESIZE, fin );    /* lines */
    }
    lines++;                      /* read another line */
    i=0;
    do {                        /* for as many lines as needed */
      for( j=0; j<PAGEWIDE && mline[i]; j++ ) {
        if( j==PAGEWIDE-2 && mline[i]!='\n' ) { /* end of screen */
          str[j++]='\\';
          str[j++]='\n';
          break;
        }
        str[j]=mline[i++];
      }
      str[j]=0;                   /* terminate */
      printf( "%s", str );
      plines++;                   /* printed another line */
    } while ( mline[i] );
  }                               /* for plines */

  fseek( fin, cpos, SEEK_SET );   /* return to last postion */
  i = mlines-lines;
  if( i<0 ) i=0;
  return( i );
}



/* ---------------------------------------------------------------------
/* interact - this is a major routine, which allows the user to view the
/* vote and make changes to it.
/* Returns RETRY_C, RETRYB_C, CONTINUE_C, DISCARD_C, or ADDROK_C
/* -----------------------------------------------------------------
*/
int interact( addrchk )
int addrchk;                       /* if true, do suspicious address */
{
  static char c, vstr[MAXGROUPS+1];
  static char name2[80], addr2[80];
  static int votes2[MAXGROUPS], mvotes2[MAXGROUPS];
  int lnum, lleft;                /* line number */
  int doneloop, i, nc, ac, vc;

  if( !dointer ) return( CONTINUE_C ); /* not in interactive mode */

  /* ask if we want to delve into it */
  do {
    if( addrchk ) {
      printf( "*** A)ddress OK, C)ontinue, D)iscard, E)xamine [E]? " );
    } else {
      printf( "*** C)ontinue, D)iscard, E)xamine [E]? " );
    }
    fgets( uin, 80, stdin );
    stripstr( uin );
    c = upcase( uin[0] );
    if( !c ) c='E';
    if( c == 'A' && addrchk ) {
      printf( "\n" );
      return( ADDROK_C );            /* continue */
    }
    if( c =='C' ) {
      printf( "\n" );
      return( CONTINUE_C );            /* continue */
    }
    if( c == 'D' ) {
      printf( " This doesn't record any result or send an ack.\n" );
      printf( " Enter YES to actually do this: " );
      fgets( uin, 80, stdin );
      if( strlook( "YES", uin, 3 ) == 0 ) {
        return( DISCARD_C );
      }
    }
  } while( c!='E' );

  printf( "%s %s\n", name, addr );
  strcpy( name2, name );          /* save current statuses */
  strcpy( addr2, addr );
  printf( "%s %s\n", name2, addr2 );
  for( i=0; i<GROUPS; i++ ) {
    votes2[i]=votes[i];
    mvotes2[i]=mvotes[i];
  }

  lnum = 1;                       /* start of message */
  for( ;; ) {
#ifdef CLRSCRN
    puts( CLRSCRN );              /* clear screen */
#else
    puts( "\n\n\n" );
#endif
    lleft = showmsg( lnum, PAGELINES ); /* show message */
    for( i=0; i<PAGEWIDE-1; uin[i++]='-' ); /* separator */
    uin[i]=0;
    if( ( i=lnum+PAGELINES ) > mlines+1 ) i = mlines+1;
    sprintf( &uin[5], "[%3d%%]", ((mlines-lleft)*100)/mlines );
    for( i=0; uin[i]; i++ );
    uin[i]='-';
    printf( "%s\n", uin );
    votes_str( votes2, mvotes2, vstr );
    printf( "A: %s\nN: %s\nV: %s\n", addr2, name2, vstr );
    for( doneloop=FALSE; !doneloop; ) {
      printf( "*** Q)uit, S)ave;  A)ddress, N)ame, V)ote;  D)own, U)p [D] ? " );
      fgets( uin, 80, stdin );
      stripstr( uin );
      c = upcase( uin[0] );
      if( !c ) c = 'D';      /* default */
      switch( c ) {
        case 'Q':                 /* quit without saving */
          printf( "\n" );
          return( RETRY_C );         /* look again */
          break;
        case 'S':                 /* save and exit */
          printf( "\n" );
          if( ac = strcmp( addr, addr2 ))
            if( ftrace ) fprintf( ftrace,"   New addr: %s\n", addr2 );
          if( nc = strcmp( name, name2 ))
            if( ftrace ) fprintf( ftrace,"   New name: %s\n", name2 );
          for( vc=FALSE, i=0; i<GROUPS; i++ ) {
            if( votes[i]!=votes2[i] || mvotes[i]!=mvotes2[i] ) {
              vc=TRUE;
              if( ftrace ) fprintf( ftrace,"   New vote: %s\n", vstr );
              break;
            }
          }
          if( ac || nc || vc ) {  /* something changed */
            strcpy( comment, "Votetaker set " );
            if( ac ) strcat( comment, "address " );
            if( nc ) strcat( comment, "name " );
            if( vc ) strcat( comment, "vote " );
          }
          strcpy( name, name2 );  /* use new values */
          strcpy( addr, addr2 );
          for( i=0; i<GROUPS; i++ ) {
            votes[i]=votes2[i];
            mvotes[i]=mvotes2[i];
          }
          return( RETRYB_C );     /* look again, force ballot found */
          break;
        case 'D':                 /* go down page */
        case '+':
        case '=':
          i = ( c=='D' ) ? PAGELINES : 1;
          if( i>lleft ) i=lleft;
          lnum += i;
          doneloop=TRUE;
          break;
        case 'U':                 /* go up page */
        case '-':
          lnum-= ( c=='U' ) ? PAGELINES : 1;
          if( lnum<1 ) lnum = 1;
          doneloop=TRUE;
          break;
        case 'A':                 /* change address */
          printf( "Enter new address: " );
          fgets( uin, 80, stdin );
          stripstr( uin );
          if( uin[0] ) strcpy( addr2, uin );
          doneloop=TRUE;
          break;
        case 'N':                 /* change address */
          printf( "Enter new name: " );
          fgets( uin, 80, stdin );
          stripstr( uin );
          if( uin[0] ) strcpy( name2, uin );
          doneloop=TRUE;
          break;
        case 'V':                 /* change votes */
          for( i=0; i<GROUPS; i++ ) {
            for( ;; ) {
              if( u_stv ) {
                printf( "  %s (1-%d/A/C) [%c]? ", groupstxt[i], GROUPS, vstr[i] );
              } else {
                printf( "  %s (Y/N/A/C) [%c]? ", groupstxt[i], vstr[i] );
              }
              fgets( uin, 80, stdin );
              stripstr( uin );
              c = upcase( uin[0] );
              if( !c ) {
                break;
              } else if ( !u_stv && c=='N' ) {
                votes2[i] = NO_V;
                mvotes2[i]=1;
                break;
              } else if ( !u_stv && c=='Y' ) {
                votes2[i] = YES_V;
                mvotes2[i]=1;
                break;
              } else if ( c=='A' || c=='-' ) {
                votes2[i] = ABSTAIN_V;
                mvotes2[i]=1;
                break;
              } else if ( c=='C' ) {
                votes2[i] = CANCEL_V;
                mvotes2[i]=1;
                break;
              } else if ( u_stv && c>='1' && c<=('0'+GROUPS) ) {
                votes2[i] = c;
                mvotes2[i] = 1;
                break;
              }
            }
          }
          doneloop=TRUE;
          break;
      }                           /* switch QS; ANV; DU */
    }                             /* for !doneloop */
  }                               /* for ( ;; ) */
}



/* ---------------------------------------------------------------------
/* patterncheck - recursive!  check to see if the first string (containing
/* wildcards) equals the 2nd string.  Returns true if it does.
/* -----------------------------------------------------------------
*/
int patterncheck( wild, txt )
char *wild;
char *txt;
{
  int i, j, j2;

  i= j = 0;                       /* start at first position */
  while( wild[i] && txt[j] ) {    /* until end of string */
    switch( wild[i] ) {
      case '?':                   /* wildcard one character */
        i++; j++;
        break;
      case '*':                   /* wildcard n characters*/
        for( j2=j; txt[j2]; j2++ ) { /* check against each */
          if( patterncheck( &wild[i+1], &txt[j2] )) return( TRUE );
        }
        if( patterncheck( &wild[i+1], &txt[j2] )) return( TRUE );
        return( FALSE );          /* sorry */
        break;
      default:                    /* character to character */
        if( ( wild[i++] | 0x20 ) != ( txt[j++] | 0x20 ) ) {
          return( FALSE );        /* failed comparison */
        }
        break;
    }  /* switch */
  }  /* while */
  while( wild[i] == '*' ) i++;    /* skip ending wildcards */
  return( !wild[i] && !txt[j] );  /* equal iff both at end */
}


/* ---------------------------------------------------------------------
/* checkaddr - check address against list of suspicious addresses.
/* Returns -1 if doesn't match any, otherwise index into address list.
/* -----------------------------------------------------------------
*/
int checkaddr()
{
  int i;

  for( i=0; i<ATXTS; i++ ) {      /* once for each wildcard */
    if( patterncheck( addrstxt[i], addr ) ) { /* recursive pattern check */
      return( i );
    }
  }
  return( -1 );
}


/* -----------------------------------------------------------------
/* finalprocess - check the vote for errors, allow correction, exit
/* when done.  Return TRUE if this should be called again, FALSE if not
/* -----------------------------------------------------------------
*/
int finalprocess( fmsg )
FILE *fmsg;                       /* name of message file */
{
  static char tmp[80];
  static int i, j, k;

  /* ---------------- check for address ---------------------------- */
  if( ( k = checkaddr()) != -1 ) {
    if( dotrace) fprintf( ftrace,"   ERROR - suspicious address (%s)!\n", addrstxt[k] );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - Suspicious address (%s)!\n", addrstxt[k] );

    switch( interact( TRUE ) ) {
      case ADDROK_C:
        break;                    /* just keep going */
      case RETRYB_C:
      case RETRY_C:
        return( TRUE );             /* try again */
        break;
      case CONTINUE_C:
        writevote( "! Invalid address" );
        if ( u_dovoteack ) {
          ackmsg( fmsg, "[bad account]" );
        }
      case DISCARD_C:
        return( FALSE );            /* all done */
    }
  }

  /* ---------------- check for no ballot -------------------------- */
  if( state != BALLOT_S ) {
    if( dotrace) fprintf( ftrace,"   ERROR - Couldn't Find Ballot!\n" );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - Couldn't Find Ballot!\n" );
    switch( interact( FALSE ) ) {
      case RETRYB_C:
        state = BALLOT_S;
      case RETRY_C:
        return( TRUE );             /* try again */
        break;
      case CONTINUE_C:
        writevote( "! No ballot" );
        if ( u_dovoteack ) {
          ackmsg( fmsg, "[no ballot]" );
        }
      case DISCARD_C:
        return( FALSE );            /* all done */

    }
  }

  /* ------------------ check for cancel --------------- */
  for( i=0, j=0; i<GROUPS; i++ ) {
    if( votes[i] == CANCEL_V ) j++;
  }
  if( j != 0 ) {
    if( dotrace) fprintf( ftrace,"   CANCEL vote\n" );
    writevote( "* Cancel" );
    if ( u_dovoteack ) {
      ackmsg( fmsg, "[canceled]" );
    }
    return( FALSE );              /* all done */
  }


  /* ------------------ check for no votes --------------- */
  for( i=0, j=0; i<GROUPS; i++ ) {
    j+= mvotes[i];
  }
  if( j == 0 ) {
    if( dotrace) fprintf( ftrace,"   ERROR - No votes!\n" );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - No votes!\n" );
    switch( interact( FALSE ) ) {
      case RETRY_C:
      case RETRYB_C:
        return( TRUE );             /* try again */
        break;
      case CONTINUE_C:
        writevote( "! No vote statement in message" );
        if ( u_dovoteack ) {
          ackmsg( fmsg, "[no votes]" );
        }
      case DISCARD_C:
        return( FALSE );              /* all done */
    }
  }

  /* ----------------- check for multiple votes --------------------- */
  for( i=0, j=FALSE; i<GROUPS; i++ ) {
    if( mvotes[i] >1 ) {
      j=TRUE; break;
    }
  }
  if( j ) {
    if( dotrace) fprintf( ftrace,"   ERROR - Conflicting votes on group(s)!\n" );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - conflicting votes on group(s)!\n" );
    for( i=0; i<GROUPS; i++ ) {   /* print offending groups */
      if( mvotes[i] >1 ) printf( "     %s\n", groupstxt[i] );
    }
    switch( interact( FALSE ) ) {
      case RETRY_C:
      case RETRYB_C:
        return( TRUE );             /* try again */
        break;
      case CONTINUE_C:
        writevote( "! Conflicting votes" );
        if ( u_dovoteack ) {
          ackmsg( fmsg, "[multiple vote]" );
          for( i=0; i<GROUPS; i++ ) {
            if( mvotes[i] >1 ) fprintf( fmsg, "    %s\n", groupstxt[i] );
          }
        }
      case DISCARD_C:
        return ( FALSE );             /* all done */
    }
  }
  /* ----------------- check for good ordering STV --------------------- */
  if( u_stv ) {
    int votes2[MAXGROUPS], i1;

    for( i=0; i<GROUPS; i++ ) votes2[i] = votes[i];
    for( i=0; i<GROUPS; i++ )
      for( j=0; j<GROUPS-1; j++ )
        if( votes2[j]>votes2[j+1] ) {
          k = votes2[j]; votes2[j] = votes2[j+1]; votes2[j+1] = k;
        }

/*    printf( " === " );
    for( i=0; i<GROUPS; i++ ) putch( votes2[i] );
    printf( "\n" ); */


    j = FALSE;  i1='1';
    for( i=0; i<GROUPS; i++ ) {
      if( votes2[i] == CANCEL_V ) {
        j = FALSE; break;
      } else if ( votes2[i] == ABSTAIN_V ) {
        ;
      } else if( votes2[i] == i1 ) {
        i1++;
      } else {
        j = TRUE; break;
      }
    }
  }

  if( j ) {
    if( dotrace) fprintf( ftrace,"   ERROR - Bad STV ordering!\n" );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - Bad STV ordering!\n" );
    switch( interact( FALSE ) ) {
      case RETRY_C:
      case RETRYB_C:
        return( TRUE );             /* try again */
        break;
      case CONTINUE_C:
        writevote( "! Bad STV ordering" );
        if( u_dovoteack ) {
          ackmsg( fmsg, "[bad stv]" );
          if ( u_votemark ) {
            fprintf( fmsg, "                                          " );
            for( k=0; u_votemark[k]; k++ )
              putc( u_votemark[k] == '%' ? vmark : u_votemark[k], fmsg );
            fputc( '\n', fmsg );
          }
          fprintf( fmsg, "Voter address: %s\n", addr );
          fprintf( fmsg, "Voter name:    %s\n", name );
          printvote( fmsg );
        }
      case DISCARD_C:
        return ( FALSE );             /* all done */
    }
  }

  /* ----------------- check for rules violations --------------------- */
  j = rule_check();
  if( j != -1 ) {
    if( dotrace) fprintf( ftrace,"   ERROR - Ballot violates rule #%d!\n", j+1 );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - Ballot violates rule #%d!\n", j+1 );
    switch( interact( FALSE ) ) {
      case RETRY_C:
      case RETRYB_C:
        return( TRUE );             /* try again */
      case CONTINUE_C:
        sprintf( tmp, "! Violated rule %d\n", j+1 );
        writevote( tmp );
        if ( u_dovoteack ) {
          ackmsg( fmsg, "[invalid case]" );
          fprintf( fmsg, "\nHere's the rule we think you violated:\n\n" );
          rule_print( j, fmsg );
        }
      case DISCARD_C:
        return ( FALSE );           /* all done */
    }
  }

  /* ----------------- check for presence of name --------------------- */
  if ( u_nametext ) {
    if ( u_needname ) {
      if( name[0]=='\0' ) {
        if( dotrace) fprintf( ftrace,"   ERROR - No name provided!\n" );
        printf( "%s (%s)\n", addr, name );
        printf( "   ERROR - No name provided!\n" );
        switch( interact( FALSE ) ) {
          case RETRY_C:
          case RETRYB_C:
            return( TRUE );             /* try again */
          case CONTINUE_C:
            writevote( "! No name given" );
            if ( u_dovoteack ) {
              ackmsg( fmsg, "[no name]" );
            }
          case DISCARD_C:
            return( FALSE );            /* all done */
        }
      }
    }
  }

  if ( dointer && !u_needname && u_wantname ) {
    if( name[0]=='\0' ) {
      printf( "%s (%s)\n", addr, name );
      printf( "   WANTNAME - No name provided!\n" );
      switch( interact( FALSE ) ) {
        case RETRY_C:
        case RETRYB_C:
          return( TRUE );             /* try again */
        case CONTINUE_C:
        case DISCARD_C:
          break;
      }
    }
  }

  /* okay, if we made it this far, all is well in la-la-land
  /* send an actual vote acknowledgement
  */

  if ( u_dovoteack ) {
    ackmsg( fmsg, "[vote ack]" );
    if ( u_votemark ) {
      fprintf( fmsg, "                                          " );
      for( k=0; u_votemark[k]; k++ )
        putc( u_votemark[k] == '%' ? vmark : u_votemark[k], fmsg );
      fputc( '\n', fmsg );
    }
    fprintf( fmsg, "Voter address: %s\n", addr );
    fprintf( fmsg, "Voter name:    %s\n", name );
    if ( u_nametext ) {
      if ( u_needname ) {
        fprintf( fmsg, "  (Real names are required for this vote.  If this is not your\n" );
        fprintf( fmsg, "   real name, then please give it here and return this entire\n" );
        fprintf( fmsg, "   message or your vote may be designated as invalid later ).\n" );
        fprintf( fmsg, "  %s\n", u_nametext );
      }
    }
    fprintf( fmsg, "\n" );
    printvote( fmsg );
  }

  votes_str( votes, mvotes, tmp );/* convert votes to string */
  writevote( tmp );               /* save votes */

  if( dotrace ) printvote( ftrace );

  return( FALSE );                /* all done */
}


/* -----------------------------------------------------------------
/* endvote - done after a vote is processed
/* -----------------------------------------------------------------
*/
void endvote( )
{
  static char fname[20];
  static FILE *fmsg;
  int i, j, k, doquote;

  if( state == NOTHING_S ) {
    return;        /* nothing done yet... */
  }

  /* find correct date of the ones we found ( if any ) */
  for( i=0; i<DTXTS; i++ ) {
    if( datefnd[i][0] ) {
      strcpy( date, datefnd[i] );
      break;
    }
  }


  /* Can't do anything with no address!
  */
 while ( addr[0] == 0 ) {
    if( dotrace) fprintf( ftrace,"   ERROR - No address!\n" );
    printf( "%s (%s)\n", addr, name );
    printf( "   ERROR - No address!\n" );
    switch( interact( FALSE ) ) {
      case CONTINUE_C:
        writevote( "! No address" );
        return;
      case DISCARD_C:
        return;
      default:
        break;
    }
  }

  /* check for no domain in address */
  do {
    for(i=0,j=FALSE; addr[i]; i++) {
      if( addr[i] == '@' ) {
        j=TRUE; break;
      }
    }
    if( j == FALSE) {
      if( dotrace) fprintf( ftrace,"   WARNING - address has no '@'!\n" );
      printf( "%s (%s)\n", addr, name );
      printf( "   WARNING - address has no '@'!\n" );
      switch( interact( FALSE ) ) {
        case CONTINUE_C:
          j = TRUE;
          break;
        case DISCARD_C:
          return;
        default:
          break;
      }
    }
  } while ( j == FALSE );

  /* First, generate an ack file name for this vote */
  if ( u_dovoteack ) {
    sprintf( fname, "acq.%c%c%c",  (count/( 26*26 )) +'a',
      (( count/26 ) % 26 ) +'a', ( count % 26 ) + 'a' );

    if ( !( fmsg = fopen( fname, UV_OPEN_WRITE ))) {
      printf( "Can't open %s for output\n", fname );
      do_exit( 2 );
    }

    /* now make a command to mail this thing */
    for( i=0, j=0; MAILCMD[i]; i++ ) {
      if( MAILCMD[i] != '#' ) {
        mline[j++] = MAILCMD[i];
      } else {
        doquote = 0;
        switch ( MAILCMD[++i] ) {
          case 'S':
            doquote = 1;
            mline[j++] = '"';
          case 's':
            for( k=0; u_voteacksub[k]; k++ ) mline[j++] = u_voteacksub[k];
            if( doquote ) mline[j++] = '"';
            break;
          case 'V':
            doquote = 1;
            mline[j++] = '"';
          case 'v':
            for( k=0; addr[k]; k++ ) {
              word l;
              for( l=0; l<strlen( ADDBADCHAR ); l++ ) {
                if( addr[k] == ADDBADCHAR[l] ) {
                  mline[j++] = '\\'; break;
                }
              }
              mline[j++] = addr[k];
            }
            if( doquote ) mline[j++] = '"';
            break;
          case 'f':
            for( k=0; fname[k]; k++ ) mline[j++] = fname[k];
            break;
          default:
            printf( "  Error in mail text string:\n   %s\n", MAILCMD );
            fprintf( fvotes, "V: ! Bad mail string\n" );
            do_exit( 12 );
            break;
        }  /* switch */
      }  /* # */
    }  /* for */
    mline[j++]='\0';
    fprintf( fcmd, "%s\n", mline );
  }

  /* keep checking the vote until we print it out */
  while ( finalprocess( fmsg ) );

  if ( u_dovoteack ) {
    ackmsg( fmsg, "[signature]" );
    fclose( fmsg );
  }
}


/* -----------------------------------------------------------------
/* checkballot - check the line for the ballot marker
/* return TRUE if it was found
/* -----------------------------------------------------------------
*/
int checkballot()
{
  int i, j;

  if ( u_votemark ) {
    for( i=0; line[i]; i++ ) {
      for( j=0; j<sizeof( u_votemark )-1; j++ ) {
        if( ( u_votemark[j] != '%' ) && ( u_votemark[j] != line[i+j] ) ) {
          break;
        }
        if( u_votemark[j] == '%' ) vmark = line[i+j]; /* votemark found */
      } /* for j */
      if( j == sizeof( u_votemark )-1 ) {
        return( TRUE );
      }
    } /* for i */
    return( FALSE );
  } else {
    return( TRUE );
  }
}


/* -----------------------------------------------------------------
/* checkvote - check a line line for voting info
/* -----------------------------------------------------------------
*/
void checkvote( inheader )
int inheader;
{
  int len, i, j, v, nword;

  /* now looking in ballot */
  len = strlen( line );

  if ( u_nametext ) {
    if( !inheader) {
      for( i=0; i<(int)( len-strlen( u_nametext )); i++ ) {
        if( strlook( u_nametext, &line[i], strlen( u_nametext )) == 0 ) {
          i+= strlen( u_nametext );     /* skip past u_nametext */
          while( line[i] && !isletter( line[i] )) i++;
          if( !line[i] ) {            /* end of line */
            break;
          }
          j = 0;                      /* now copy to name, compress spaces */
          while( line[i] && line[i] != '\n' && j<(sizeof(name)-8) ) {
            if( line[i] == ' ' || line[i] == '\t' ) {
              name[j++]=' ';
              while( line[i] == ' ' || line[i] == '\t' ) i++;
            } else {
              name[j++]=line[i++];
            }
          }
          name[j++]='\0';
          if( dotrace) fprintf( ftrace, "  Name Found: %s\n", name );
        }
      }
    }
  }


  if ( u_explicitvote ) {
    for( i=0, j=FALSE; i<len; i++ ) {
      if( strlook( "vote", &line[i], 4 ) == 0 ) {
        j = TRUE; break;
      }
    }
    if( !j ) return;
  }

  for( v=0; v<GROUPS; v++ ) {     /* check for each group */
    for( i=0; i<len-groupslen[v]; i++ ) {
      if( !isletter( line[i] ) ) continue;
      if( strlook( groupstxt[v], &line[i], groupslen[v] ) == 0 ) {
        if (isletter(line[i+groupslen[v]] ) ) continue;
        if( line[i+groupslen[v]] == '.' && isletter( line[i+groupslen[v]+1] ))
          continue;
        /* found match! */
        if( dotrace) fprintf( ftrace, "%s", line );
        nword = 1;                /* new word */
        for( j=0; j<len; j++ ) {  /* now check for vote */
          if( isletter( line[j] ) ) {
            if( nword && ( j<i || j>(i+groupslen[v]))) {
              if( strlook( &line[j], "yes", 3 ) == 0 &&
                  !isletter( line[j+3] ) ) {
                if( votes[v] != YES_V ) mvotes[v]++;
                votes[v] = YES_V;
                break;
              }
              if( !u_stv && strlook( &line[j], "no", 2 ) == 0 &&
                  !isletter( line[j+2] ) ) {
                if( votes[v] != NO_V ) mvotes[v]++;
                votes[v] = NO_V;
                break;
              }
              if( !u_stv && strlook( &line[j], "against", 7 ) == 0 &&
                  !isletter( line[j+7] ) ) {
                if( votes[v] != NO_V ) mvotes[v]++;
                votes[v] = NO_V;
                break;
              }
              if( strlook( &line[j], "abstain", 7 ) == 0 &&
                  !isletter( line[j+7] ) ) {
                if( votes[v] != ABSTAIN_V || mvotes[v] == 0 ) mvotes[v]++;
                votes[v] = ABSTAIN_V;
                break;
              }
              if( strlook( &line[j], "cancel", 6 ) == 0 &&
                  !isletter( line[j+6] ) ) {
                if( votes[v] != ABSTAIN_V || mvotes[v] == 0 ) mvotes[v]++;
                votes[v] = CANCEL_V;
                break;
              }
              if( !u_stv && strlook( &line[j], "for", 3 ) == 0 &&
                  !isletter( line[j+3] ) ) {
                if( votes[v] != YES_V ) mvotes[v]++;
                votes[v] = YES_V;
                break;
              }
            } /* if nword */
            nword = 0;            /* not in new word */
          } else {                /* alpha */
            if( u_stv && line[j]>='0' && line[j]<='9' ) {
              if( votes[v] != line[j] ) mvotes[v]++;
              votes[v] = line[j];
              break;
            }
            nword = 1;            /* next char is start of new word */
          }
        } /* for j */
        break;
      }                           /* if group found */
    } /* for i */
  } /* for v */
}


/* ----------------------------------------------------------------------
/*  This does the tricky work of decoding which is the name and which is
/*  the address.  Used for both From: and Reply-To:
/* ----------------------------------------------------------------------
*/
void decode_nameaddr( ipos, name, addr )
int  ipos;                        /* initial position in line[] */
char *name;                       /* resulting name */
char *addr;                       /* resulting address */
{

  int i, j, baddr, deep;
  static char temp[80], c;
  int state;

#define FIND_NEXT 0
#define IN_ADDR 1
#define IN_NAME 2

  addr[0]=0; name[0]=0;           /* nothing in these yet */

  baddr = FALSE;                  /* check for bracketed address */
  for( i=ipos; line[i]; i++ ) {
    if( line[i] == '<' ) {
      baddr = TRUE;
    }
  }

  i = ipos;                       /* start at beginning */
  deep=0;                         /* only one level deep */
  state = FIND_NEXT;
  while( line[i] ) {
    c = line[i++];
    if( j == sizeof(temp)-3 ) c = '\n';  /* too damn long! */
    if( c == '\t' ) c = ' ';      /* tabs to spaces */
    if( state != IN_NAME && c == ',' ) { /* if bozo adds multiple addresses */
      c = '\n';
      line[i] = 0;
    }
    switch( state ) {
      case FIND_NEXT:             /* looking for anything */
        j = 0;
        switch( c ) {
          case '\n':              /* end of line */
          case ' ':               /* space */
            break;
          case '(':              /* start of name */
            state = IN_NAME;
            deep = 1;
            break;
          case '<':               /* start of address */
            state = IN_ADDR;
            break;
          default:                /* other */
            if( baddr ) {         /* bracketed address exists */
              state = IN_NAME;    /* then this must be the name */
              deep = 1;
            } else {
              state = IN_ADDR;
            }
            if( c != '"' ) temp[j++] = c; /* save character */
        }
        break;

        case IN_ADDR:
          if( j==70 || c == '>' || c == '(' || c=='\n' ) { /* end of address */
            while( j>0 && temp[j-1] == ' ' ) j--; /* erase any end spaces */
            temp[j] = '\0';       /* terminate string */
            temp[70] = '\0';      /* just in case... */
            strcpy( addr, temp ); /* and copy */
            state = FIND_NEXT;
            if( c=='(' ) i--;
          } else {
            temp[j++]=c;
          }
          break;

        case IN_NAME:
          if( c==')') --deep;
          if( j==70 | deep == 0 || c == '<' || c=='\n' ) {  /* end of name */
            deep = 0;
            while( j>0 && temp[j-1] == ' ' ) j--; /* erase any end spaces */
            if( temp[j-1] == '"' ) j--; /* strip last quote */
            temp[j] = '\0';        /* terminate string */
            temp[70] = '\0';       /* just in case... */
            for( j=0; temp[j++]==' '; ); /* skip initial spaces */
            strcpy( name, &temp[j-1] ); /* and copy */
            state = FIND_NEXT;
            if( c=='<' ) i--;
          } else {
            if(c=='(') deep++;
            if( j!=0 || c!='"' ) { /* strip initial quote */
              temp[j++]=c;
            }
          }
          break;

    }                             /* switch */
  }                               /* while line[i] */
  for(i=0; addr[i]; i++) {        /* no spaces in addresses, stupid voters! */
    if( addr[i] == ' ' ) addr[i] = 0;
  }
}


/* ----------------------------------------------------------------------
/*  Handle Reply-To: field
/* ----------------------------------------------------------------------
*/
void replyto()
{
  static char name2[80];
  static char addr2[80];

  if( dotrace) fprintf( ftrace, "\n\n%s", line );

  /* decode the name and address */
  decode_nameaddr( 10, name2, addr2 );

  if( name2[0] != 0 ) {           /* if a name, copy it */
    strcpy( name, name2 );
  }
  if( addr2[0] != 0 ) {           /* if an address, copy it */
    strcpy( addr, addr2 );
  }

  if( dotrace ) {
    fprintf( ftrace, "  Address: %s\n", addr );
    fprintf( ftrace, "  Name:    %s\n\n", name );
  }

}


/* ----------------------------------------------------------------------
/*  Handle date in header field
/* ----------------------------------------------------------------------
*/
void dodate( idx )
int idx;                          /* which date in array */
{
  int i, j, d, datepart;
  char c;
  static char part[80];
  static char days[14][5]= {
    "Mon,", "Tue,", "Wed,", "Thu,", "Fri,", "Sat,", "Sun,",
    "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" };
  int done=FALSE;

  i= datepart = d = 0;
  while( line[i]!='\n' ) {
    while( line[i] == ' ' || line[i]== '\t' ) i++; /* skip whitespace */
    for( j=0; line[i]!=' ' && line[i]!='\t' && line[i]!='\n'; part[j++]=line[i++] );
    part[j] = 0;

    if( datepart == 0 ) {         /* haven't found date yet */
      for( j=0; j<14; j++ ) {
        if( strcmp( part, days[j] ) == 0 ) { /* found it! */
          datepart = 1;
          break;
        }
      }
    }
    if( datepart > 0 ) {
      for( j=0; part[j]; datefnd[idx][d++]=part[j++] );
      if( ++datepart < 7 ) {
        datefnd[idx][d++]=' ';
      } else {
        break;
      }
    }
  }
  datefnd[idx][d] = 0;

}


/* ---------------------------------------------------------------------
/* Print help
/* ---------------------------------------------------------------------
*/
void help()
{
  printf( "  uvvote (-i) <votes file> <results file> (<trace file>)\n" );
  printf( "    The -i option makes error handling interactive - you will be\n" );
  printf( "      given the option to hand correct any errors in the vote.\n" );
  printf( "    The <trace file> is optional.\n\n" );
  printf( "  uvvote -t     <- just test voting rules and ack file\n\n" );
  do_exit( 1 );
}



/* ---------------------------------------------------------------------
/* Main procedure - drive all the other procedures
/* ---------------------------------------------------------------------
*/
int main( argc, argv )
int argc;                         /* number of arguments to uvvote */
char **argv;                      /* arguments to uvvote */
{
  int i, j, isstart, isend, endit, inheader;
  postype pos;                    /* position in file */
  int argp;                       /* argument position */

#define VOTESFILE argv[argp]
#define RESULTSFILE argv[argp+1]
#define TRACEFILE argv[argp+2]

  init_global_config();
  rules = (struct uv_s_rules*)malloc(u_rulesmax*sizeof(struct uv_s_rules));
  for ( i=0; i<u_rulesmax; i++ ) {
    rules[i].test1 = (char*)malloc(GROUPS*sizeof(char));
    rules[i].test2 = (char*)malloc(GROUPS*sizeof(char));
  }

  printf("\n%s\n", VERSION );

  dotest = dointer = FALSE;

  if( argc == 1 ) help();
  argp = 1;                       /* argument pointer */
  if( argv[argp][0]=='-' ) {      /* option */
    switch( argv[argp][1] ) {
      case 't':                   /* test */
        dotest = TRUE;
        break;
      case 'i':                   /* interactive */
        dointer = TRUE;
        break;
      default:
        help();
    }
    argp++;
  }

  if( !dotest ) {                 /* need real arguments */
    if( argc-argp == 3 ) {
      dotrace = TRUE;
    } else if ( argc-argp == 2 ) {
      dotrace = FALSE;
    } else {
      help();
    }
  }

  lmailstart = strlen( MAILSTART );
  if( lmailstart == 0 ) {
    fprintf( stderr, "Error: MAILSTART must be set in usevote.h\n" );
  }
  lmailend = strlen( MAILEND );
  startisend = FALSE;
  if( lmailstart == lmailend ) {
    if( strncmp( MAILSTART, MAILEND, lmailstart ) == 0 ) {
      startisend = TRUE;
    }
  }

  rule_read();
  if( dotest ) {
   if( nrules == 0 ) {
      printf( "No rules found.\n\n" );
    }
    dump_config();
    if ( u_dovoteack ) {
      printf( "** Testing %s contents\n", u_acktxtfile );
      printf( "\n** Testing [vote ack]\n" );
      ackmsg( stdout, "[vote ack]" );
      printf( "\n** Testing [canceled]\n" );
      ackmsg( stdout, "[canceled]" );
      printf( "\n** Testing [bad account]\n" );
      ackmsg( stdout, "[bad account]" );
      printf( "\n** Testing [no ballot]\n" );
      ackmsg( stdout, "[no ballot]" );
      printf( "\n** Testing [no votes]\n" );
      ackmsg( stdout, "[no votes]" );
      printf( "\n** Testing [multiple vote]\n" );
      ackmsg( stdout, "[multiple vote]" );
      printf( "\n** Testing [invalid case]\n" );
      ackmsg( stdout, "[invalid case]" );
      printf( "\n** Testing [no name]\n" );
      ackmsg( stdout, "[no name]" );
      if( u_stv ) {
        printf( "\n** Testing [bad stv]\n" );
        ackmsg( stdout, "[bad stv]" );
      }
      printf( "\n** Testing [signature]\n" );
      ackmsg( stdout, "[signature]" );
      printf( "\n\n  UseVote sez: Everything looks okay...\n" );
    }

    do_exit( 0 );
  }

  if ( !( fin = fopen( VOTESFILE, UV_OPEN_READ ))) {
    printf( "Can't open %s for input\n", VOTESFILE );
    do_exit( 2 );
  }

  if ( u_dovoteack ) {
    if ( !( fcmd = fopen( u_cmdfile, UV_OPEN_WRITE ))) {
      printf( "Can't open %s for output\n", u_cmdfile );
      do_exit( 2 );
    }
#ifdef VOTEACKPREFIX
      fprintf( fcmd, "%s\n", VOTEACKPREFIX );
#endif /* VOTEACKPREFIX */
  }


  if ( ( fvotes = fopen( RESULTSFILE, UV_OPEN_READ ))) {
    fclose( fvotes );
    fprintf( stderr, "\nWARNING! Results file %s already exists.\n", RESULTSFILE );
    fprintf( stderr, "Do you really want to write over it? " );
    fgets( uin, 80, stdin );
    stripstr( uin );
    if( upcase( uin[0] )!='Y' ) {
      do_exit( 20 );
    }
    printf( "\n" );
  }

  if ( !( fvotes = fopen( RESULTSFILE, UV_OPEN_WRITE ))) {
    printf( "Can't open %s for output\n", RESULTSFILE );
    do_exit( 2 );
  }

  if( dotrace ) {
    if ( !( ftrace = fopen( TRACEFILE, UV_OPEN_WRITE ))) {
      printf( "Can't open %s for output\n", TRACEFILE );
      do_exit( 2 );
    }
  }

  count = 0;                      /* no votes yet */
  for( i=0; i<GROUPS; i++ ) {     /* find group name lengths */
    for( j=0; groupstxt[i][j]; j++ ) {
      if( space( groupstxt[i][j] )) break;
    }
    groupslen[i] = j;
  }
  for( i=0; i<DTXTS; i++ ) {     /* find date string lengths */
    dateslen[i] = strlen( datestxt[i] );
  }


  /* ----------------------------------------------------------------
  /*  This is the big loop  -  Once per line
  /* ----------------------------------------------------------------
  */
  do {

    pos = ftell( fin );
    if( endit = !fgets( line, LINESIZE, fin ) ) {
      isend = TRUE; isstart = FALSE;
      strcpy( line, "[END OF FILE]" );
    } else {
      isend = !strncmp( line, MAILEND, lmailend ) && lmailend;
      isstart = !strncmp( line, MAILSTART, lmailstart );
    }
    mlines++;

    if( startisend && isstart ) { /* end or start if they're the same? */
      if( state == NOTHING_S ) {
        isstart = TRUE; isend = FALSE;
      } else {
        isstart = FALSE; isend = TRUE;
      }
    }

    if( lmailend == 0 && isstart ) {
      isend = TRUE;
    }

/*    printf( "%s", line );
    printf( "  isstart: %d   isend: %d\n", isstart, isend );
*/

    if( isend ) {
      if( state == NEEDNAME_S ) {
        count++;
        printf( "  Error!  Message found with no From: before\n" );
        printf( "    %s\n", line );
        if( dotrace) fprintf( ftrace,"   ERROR - Message found with no From: before\n" );
        if( dotrace) fprintf( ftrace,"     %s\n", line );
#ifdef VOTEPAUSE
        getc( stdin );
#endif
      } else {
        if( state != NOTHING_S ) {
          endvote();                /* finish last ballot */
          count++;
        }
      }
      state = NOTHING_S;
    }

    /* found a new vote */
    if( isstart ) {
      vmark='1';                  /* don't know which ballot */
      for( i=0; i<GROUPS; i++ ) {
        votes[i] = ABSTAIN_V;     /* assume abstention */
        mvotes[i] = 0;
      }
      for( i=0; i<DTXTS; i++ ) {  /* blank out dates */
        datefnd[i][0]=0;
      }
      date[0] = addr[0] = name[0] = comment[0] = 0;
      mpos = pos;                 /* starting position of message */
      mlines = 0;
      state = NEEDNAME_S;
      inheader = TRUE;            /* in header of vote */
    }

    /* one header "line" can stretch over several physical lines. Consolidate
    /* them here, since that's the elegant thing to do
    */
    if( inheader ) {
      if( line[0]=='\n' ) {        /* blank line means no more header */
        inheader = FALSE;
      } else {
        char c;
        for( i=0;; ) {            /* check for more */
          for( j=0; i<LINESIZE-2 && line[i]!='\n'; i++ ) { /* end of line */
            if( line[i]!=' ' && line[i] != '\t' ) j=1; /* all blank line */
          }
          if( j == 0 ) {          /* all blank line! */
            c = '\n';
            ungetc( c, fin );     /* push all blank line */
          } else {
            c = getc( fin );        /* peek at next line */
            ungetc( c, fin );       /* put it back */
          }
          if( c == ' ' || c == '\t' ) { /* whitespace, read next line */
            line[i] = 0;          /* just in case error... */
            endit = !fgets( &line[i], LINESIZE-i-2, fin );
            mlines++;
          } else {
            line[i++] = '\n';     /* end line */
            line[i++] = 0;
            break;
          }
        }
      }
    }

    if( state == NEEDNAME_S ) {
      if( inheader && addr[0]==0 && strncmp( line, "From:", 5 ) == 0 ) {
        if( dotrace ) fprintf( ftrace, "\n\n%s", line );
        decode_nameaddr( 6, name, addr );
        if( dotrace ) {
          fprintf( ftrace, "Address: %s\n", addr );
          fprintf( ftrace, "Name:    %s\n\n", name );
        }
        if ( u_votemark ) {
          state = NEEDBALLOT_S;
        } else {
          state = BALLOT_S;
        }
      }
    }

    if( state == NEEDBALLOT_S || state == BALLOT_S ) { /* Reply-To: ? */
      if( inheader && strncmp( line, "Reply-To:", 9 ) == 0 ) {
        replyto();
      }
    }

    /* get date if we can */
    if( inheader ) {
      for( i=0; i<DTXTS; i++ ) {
        if( datefnd[i][0] ) break;  /* don't do useless compares */
  if( strncmp( line, datestxt[i], dateslen[i] ) == 0 ) {
          dodate( i );
        }
      }
    }

    if( state == NEEDBALLOT_S ) {               /* looking for ballot */
      if( checkballot() ) {
        state = BALLOT_S;
      }
    }

    if( state == BALLOT_S || ( u_votemark && state == NEEDNAME_S )) {
      checkvote( inheader );   /* check line for voting info */
    }

/*    {
      char stxt[4][20] = {"NOTHING_S", "NEEDNAME_S", "NEEDBALLOT_S", "BALLOT_S"};
      printf( "  State: %s\n\n", stxt[state] );
    }
*/


  } while ( !endit );

  if ( u_dovoteack ) {
#ifdef VOTEACKSUFFIX
    fprintf( fcmd, "%s\n", VOTEACKSUFFIX );
#endif
  }

  printf( "\nAll Done... %d votes processed.\n", count );

  do_exit(0);
  return( 0 );
}
